import time
import logging
from threading import Thread
from servicebus.command import cmd


DIDA_TIMEOUT = 60
JIFFY = 5


class PingWatcher:
    def __init__(self, receiver):
        self.run = True
        self.receiver = receiver

    def do_watch(self):
        ip = self.receiver.host
        ret = self.ping(ip)
        if not ret and self.receiver.connected:
            self.receiver.connected = False
            logging.info("Ping Error!")
            return True
        duration = 0
        while self.run:
            time.sleep(JIFFY)
            duration += JIFFY
            if duration >= DIDA_TIMEOUT:
                duration = 0
                ret = self.ping(ip)
                if not ret and self.receiver.connected:
                    self.receiver.connected = False
                    logging.info("Ping Error!")
                    return True
                if not self.receiver.connected:
                    logging.info("Message receiver exited, PingWatcher out.")
                    return True
        return False

    def stop(self):
        logging.info("PingWatcher Killed")
        self.run = False

    def do_ping(self, ip):
        command_linux = "ping -c 1 %s | grep \" 1 received,\"" % (ip)
        command_macosx = "ping -c 1 %s | grep \" 1 packets received,\"" % (ip)
        for command in [command_linux, command_macosx]:
            ret = cmd(command)[0]
            if ret != "":
                return True
        return False

    def ping(self, ip):
        for i in range(3):
            ret = self.do_ping(ip)
            if ret:
                return True
            time.sleep(1)
        return False

    def run_watch(self):
        while self.run:
            try:
                if self.do_watch():
                    return
                else:
                    if not self.run:
                        logging.info("PingWatcher out")
                        return
            except Exception as e:
                logging.error(e)

    def join(self):
        self.thread.join()

    @classmethod
    def start_watch(cls, receiver):
        pw = PingWatcher(receiver)
        thread = Thread(target=pw.run_watch)
        thread.start()
        pw.thread = thread
        return pw



import threading
import logging
import asyncore


close_sockets_lock = threading.Lock()


def _do_close_sockets():
    for value in asyncore.socket_map.values():
        logging.info("Close Socket: %s" % str(value))
        value.socket.close()
    logging.info("Clean socket map!")
    asyncore.socket_map = {}


def close_sockets():
    global close_sockets_lock
    try:
        close_sockets_lock.acquire()
        _do_close_sockets()
    finally:
        close_sockets_lock.release()


def num_sockets():
    global close_sockets_lock
    try:
        close_sockets_lock.acquire()
        return len(asyncore.socket_map.values())
    finally:
        close_sockets_lock.release()



import json
import hashlib
from datetime import datetime, timedelta
from xml.dom.minidom import parseString
from servicebus.event import Event


def do_generate_token(configuration, date=None):
    key = configuration.secret_token
    td = timedelta(1)
    if date is None:
        datestr = datetime.now().isoformat()[:10]
    elif date == "prev":
        datestr = (datetime.now() - td).isoformat()[:10]
    elif date == "next":
        datestr = (datetime.now() + td).isoformat()[:10]
    token_str = "%s - %s" % (key, datestr)
    return hashlib.sha1(token_str.encode()).hexdigest()


class AbstractMessageParser(object):
    def set_configuration(self, configuration):
        self.configuration = configuration

    def generate_token(self, date=None):
        return do_generate_token(self.configuration, date)


class XmlParserHelper(object):
    def get_text(self, node):
        rc = []
        for child in node.childNodes:
            if child.nodeType == node.TEXT_NODE or child.nodeType == node.CDATA_SECTION_NODE:
                rc.append(child.data)
        return ''.join(rc)


class JSONParamsParser(XmlParserHelper):
    def __init__(self):
        self.params = {}

    def parse(self):
        params_node = self.params_node
        self.parse_param(params_node)

    def set_params_node(self, value):
        self.params_node = value

    def parse_param(self, params_node):
        json_code = self.get_text(params_node)
        self.params = json.loads(json_code)


# XML message format:
#   <?xml version="1.0"?>
#   <event version="1">
#       <id>EVENT_ID</id>
#       <token>EVENT_TOKEN</token>
#       <catgory>EVENT_CATEGORY</catgory>
#       <service>SERVICE_NAME</service>
#       <params>
#           JSON_PARAMS
#       </params>
#   </event>
class XmlMessageParser(AbstractMessageParser, XmlParserHelper):
    def parse(self, xml_doc):
        try:
            doc = parseString(xml_doc)

            if not self.validate_xml(doc):
                return None
            root = doc.childNodes[0]
            version = self.get_message_version(root)
            eid = self.get_event_id(root)
            category = self.get_category(root)
            service = self.get_service(root)
            token = self.get_token(root)

            params_parser = JSONParamsParser()
            params_parser.set_params_node(root.getElementsByTagName('params')[0])
            params_parser.parse()
            params = params_parser.params

            return Event(eid, category, service, token, params, version)
        except Exception:
            # if got any exception in parse return None
            return None

    def validate_xml(self, doc):
        root = doc.childNodes[0]
        if not root.tagName == 'event':
            return False
        if not root.hasAttribute('version'):
            return False
        return True

    def validate_token(self, token):
        if self.generate_token() == token:
            return True
        elif self.generate_token("prev") == token:
            return True
        elif self.generate_token("next") == token:
            return True
        return False

    def get_message_version(self, root):
        return root.getAttribute('version')

    def get_event_id(self, root):
        node = root.getElementsByTagName('id')[0]
        return self.get_text(node)

    def get_category(self, root):
        node = root.getElementsByTagName('catgory')[0]
        return self.get_text(node)

    def get_service(self, root):
        node = root.getElementsByTagName('service')[0]
        return self.get_text(node)

    def get_token(self, root):
        node = root.getElementsByTagName('token')[0]
        return self.get_text(node)


class XmlResponseParser(XmlParserHelper):
    def parse(self, xml_doc):
        try:
            doc = parseString(xml_doc)
            root = doc.childNodes[0]
            rid = self.get_request_id(root)
            message = self.get_message(root)
            return (rid, message)
        except Exception:
            # if got any exception in parse return None
            return None

    def get_request_id(self, root):
        node = root.getElementsByTagName('id')[0]
        return self.get_text(node)

    def get_message(self, root):
        node = root.getElementsByTagName('message')[0]
        return self.get_text(node)


ID_SEED = 0
MESSAGE_TEMPLATE = """<?xml version="1.0"?>
<event version="1">
    <id>%s</id>
    <token>%s</token>
    <catgory>%s</catgory>
    <service>%s</service>
    <params><![CDATA[%s]]></params>
</event>
"""
REPORT_XML_TEMPLATE = """<?xml version="1.0"?>
<response>
    <id>%s</id>
    <message><![CDATA[%s]]></message>
</response>
"""


class XmlRequestGenerator(object):
    def __init__(self, configuration, category, service, message):
        self.message = message
        self.configuration = configuration
        self.category = category
        self.service = service
        self.message = message

    def generate_id(self):
        global ID_SEED
        ID_SEED = ID_SEED + 1
        return ID_SEED

    def generate_token(self):
        return do_generate_token(self.configuration)

    def encode_params(self):
        return json.dumps(self.message)

    def to_xml(self):
        return MESSAGE_TEMPLATE % (
            self.generate_id(),
            self.generate_token(),
            self.category, self.service, self.encode_params())


class XmlResponseGenerator(object):
    def __init__(self, eid, message):
        self.eid = eid
        self.message = message

    def to_xml(self):
        return REPORT_XML_TEMPLATE % (self.eid, self.message)



"""Using Pika with a Twisted reactor.

Supports two methods of establishing the connection, using TwistedConnection
or TwistedProtocolConnection. For details about each method, see the docstrings
of the corresponding classes.

The interfaces in this module are Deferred-based when possible. This means that
the connection.channel() method and most of the channel methods return
Deferreds instead of taking a callback argument and that basic_consume()
returns a Twisted DeferredQueue where messages from the server will be
stored. Refer to the docstrings for TwistedConnection.channel() and the
TwistedChannel class for details.

"""
import functools
from twisted.internet import defer, error, reactor
from twisted.python import log

from servicebus.pika import exceptions
from servicebus.pika.adapters import base_connection


class ClosableDeferredQueue(defer.DeferredQueue):
    """
    Like the normal Twisted DeferredQueue, but after close() is called with an
    Exception instance all pending Deferreds are errbacked and further attempts
    to call get() or put() return a Failure wrapping that exception.
    """

    def __init__(self, size=None, backlog=None):
        self.closed = None
        super(ClosableDeferredQueue, self).__init__(size, backlog)

    def put(self, obj):
        if self.closed:
            return defer.fail(self.closed)
        return defer.DeferredQueue.put(self, obj)

    def get(self):
        if self.closed:
            return defer.fail(self.closed)
        return defer.DeferredQueue.get(self)

    def close(self, reason):
        self.closed = reason
        while self.waiting:
            self.waiting.pop().errback(reason)
        self.pending = []


class TwistedChannel(object):
    """A wrapper wround Pika's Channel.

    Channel methods that normally take a callback argument are wrapped to
    return a Deferred that fires with whatever would be passed to the callback.
    If the channel gets closed, all pending Deferreds are errbacked with a
    ChannelClosed exception. The returned Deferreds fire with whatever
    arguments the callback to the original method would receive.

    The basic_consume method is wrapped in a special way, see its docstring for
    details.
    """

    WRAPPED_METHODS = ('exchange_declare', 'exchange_delete', 'queue_declare',
                       'queue_bind', 'queue_purge', 'queue_unbind', 'basic_qos',
                       'basic_get', 'basic_recover', 'tx_select', 'tx_commit',
                       'tx_rollback', 'flow', 'basic_cancel')

    def __init__(self, channel):
        self.__channel = channel
        self.__closed = None
        self.__calls = set()
        self.__consumers = {}

        channel.add_on_close_callback(self.channel_closed)

    def channel_closed(self, channel, reply_code, reply_text):
        # enter the closed state
        self.__closed = exceptions.ChannelClosed(reply_code, reply_text)
        # errback all pending calls
        for d in self.__calls:
            d.errback(self.__closed)
        # close all open queues
        for consumers in self.__consumers.values():
            for c in consumers:
                c.close(self.__closed)
        # release references to stored objects
        self.__calls = set()
        self.__consumers = {}

    def basic_consume(self, *args, **kwargs):
        """Consume from a server queue. Returns a Deferred that fires with a
        tuple: (queue_object, consumer_tag). The queue object is an instance of
        ClosableDeferredQueue, where data received from the queue will be
        stored. Clients should use its get() method to fetch individual
        message.
        """
        if self.__closed:
            return defer.fail(self.__closed)

        queue = ClosableDeferredQueue()
        queue_name = kwargs['queue']
        kwargs['consumer_callback'] = lambda *args: queue.put(args)
        self.__consumers.setdefault(queue_name, set()).add(queue)

        try:
            consumer_tag = self.__channel.basic_consume(*args, **kwargs)
        except Exception:
            return defer.fail()

        return defer.succeed((queue, consumer_tag))

    def queue_delete(self, *args, **kwargs):
        """Wraps the method the same way all the others are wrapped, but removes
        the reference to the queue object after it gets deleted on the server.

        """
        wrapped = self.__wrap_channel_method('queue_delete')
        queue_name = kwargs['queue']

        d = wrapped(*args, **kwargs)
        return d.addCallback(self.__clear_consumer, queue_name)

    def basic_publish(self, *args, **kwargs):
        """Make sure the channel is not closed and then publish. Return a
        Deferred that fires with the result of the channel's basic_publish.

        """
        if self.__closed:
            return defer.fail(self.__closed)
        return defer.succeed(self.__channel.basic_publish(*args, **kwargs))

    def __wrap_channel_method(self, name):
        """Wrap Pika's Channel method to make it return a Deferred that fires
        when the method completes and errbacks if the channel gets closed. If
        the original method's callback would receive more than one argument, the
        Deferred fires with a tuple of argument values.

        """
        method = getattr(self.__channel, name)

        @functools.wraps(method)
        def wrapped(*args, **kwargs):
            if self.__closed:
                return defer.fail(self.__closed)

            d = defer.Deferred()
            self.__calls.add(d)
            d.addCallback(self.__clear_call, d)

            def single_argument(*args):
                """
                Make sure that the deferred is called with a single argument.
                In case the original callback fires with more than one, convert
                to a tuple.
                """
                if len(args) > 1:
                    d.callback(tuple(args))
                else:
                    d.callback(*args)

            kwargs['callback'] = single_argument

            try:
                method(*args, **kwargs)
            except Exception:
                return defer.fail()
            return d

        return wrapped

    def __clear_consumer(self, ret, queue_name):
        self.__consumers.pop(queue_name, None)
        return ret

    def __clear_call(self, ret, d):
        self.__calls.discard(d)
        return ret

    def __getattr__(self, name):
        # Wrap methods defined in WRAPPED_METHODS, forward the rest of accesses
        # to the channel.
        if name in self.WRAPPED_METHODS:
            return self.__wrap_channel_method(name)
        return getattr(self.__channel, name)


class IOLoopReactorAdapter(object):
    """An adapter providing Pika's IOLoop interface using a Twisted reactor.

    Accepts a TwistedConnection object and a Twisted reactor object.

    """

    def __init__(self, connection, reactor):
        self.connection = connection
        self.reactor = reactor
        self.started = False

    def add_timeout(self, deadline, callback_method):
        """Add the callback_method to the IOLoop timer to fire after deadline
        seconds. Returns a handle to the timeout. Do not confuse with
        Tornado's timeout where you pass in the time you want to have your
        callback called. Only pass in the seconds until it's to be called.

        :param int deadline: The number of seconds to wait to call callback
        :param method callback_method: The callback method
        :rtype: twisted.internet.interfaces.IDelayedCall

        """
        return self.reactor.callLater(deadline, callback_method)

    def remove_timeout(self, call):
        """Remove a call

        :param twisted.internet.interfaces.IDelayedCall call: The call to cancel

        """
        call.cancel()

    def stop(self):
        # Guard against stopping the reactor multiple times
        if not self.started:
            return
        self.started = False
        self.reactor.stop()

    def start(self):
        # Guard against starting the reactor multiple times
        if self.started:
            return
        self.started = True
        self.reactor.run()

    def remove_handler(self, _):
        # The fileno is irrelevant, as it's the connection's job to provide it
        # to the reactor when asked to do so. Removing the handler from the
        # ioloop is removing it from the reactor in Twisted's parlance.
        self.reactor.removeReader(self.connection)
        self.reactor.removeWriter(self.connection)

    def update_handler(self, _, event_state):
        # Same as in remove_handler, the fileno is irrelevant. First remove the
        # connection entirely from the reactor, then add it back depending on
        # the event state.
        self.reactor.removeReader(self.connection)
        self.reactor.removeWriter(self.connection)

        if event_state & self.connection.READ:
            self.reactor.addReader(self.connection)

        if event_state & self.connection.WRITE:
            self.reactor.addWriter(self.connection)


class TwistedConnection(base_connection.BaseConnection):
    """A standard Pika connection adapter. You instantiate the class passing the
    connection parameters and the connected callback and when it gets called
    you can start using it.

    The problem is that connection establishing is done using the blocking
    socket module. For instance, if the host you are connecting to is behind a
    misconfigured firewall that just drops packets, the whole process will
    freeze until the connection timeout passes. To work around that problem,
    use TwistedProtocolConnection, but read its docstring first.

    Objects of this class get put in the Twisted reactor which will notify them
    when the socket connection becomes readable or writable, so apart from
    implementing the BaseConnection interface, they also provide Twisted's
    IReadWriteDescriptor interface.

    """

    def __init__(self,
                 parameters=None,
                 on_open_callback=None,
                 on_open_error_callback=None,
                 on_close_callback=None,
                 stop_ioloop_on_close=False):
        super(TwistedConnection, self).__init__(
            parameters=parameters,
            on_open_callback=on_open_callback,
            on_open_error_callback=on_open_error_callback,
            on_close_callback=on_close_callback,
            ioloop=IOLoopReactorAdapter(self, reactor),
            stop_ioloop_on_close=stop_ioloop_on_close)

    def _adapter_connect(self):
        """Connect to the RabbitMQ broker"""
        # Connect (blockignly!) to the server
        error = super(TwistedConnection, self)._adapter_connect()
        if not error:
            # Set the I/O events we're waiting for (see IOLoopReactorAdapter
            # docstrings for why it's OK to pass None as the file descriptor)
            self.ioloop.update_handler(None, self.event_state)
        return error

    def _adapter_disconnect(self):
        """Called when the adapter should disconnect"""
        self.ioloop.remove_handler(None)
        self._cleanup_socket()

    def _handle_disconnect(self):
        """Do not stop the reactor, this would cause the entire process to exit,
        just fire the disconnect callbacks

        """
        self._on_connection_closed(None, True)

    def _on_connected(self):
        """Call superclass and then update the event state to flush the outgoing
        frame out. Commit 50d842526d9f12d32ad9f3c4910ef60b8c301f59 removed a
        self._flush_outbound call that was in _send_frame which previously
        made this step unnecessary.

        """
        super(TwistedConnection, self)._on_connected()
        self._manage_event_state()

    def channel(self, channel_number=None):
        """Return a Deferred that fires with an instance of a wrapper around the
        Pika Channel class.

        """
        d = defer.Deferred()
        base_connection.BaseConnection.channel(self, d.callback, channel_number)
        return d.addCallback(TwistedChannel)

    # IReadWriteDescriptor methods

    def fileno(self):
        return self.socket.fileno()

    def logPrefix(self):
        return "twisted-pika"

    def connectionLost(self, reason):
        # If the connection was not closed cleanly, log the error
        if not reason.check(error.ConnectionDone):
            log.err(reason)

        self._handle_disconnect()

    def doRead(self):
        self._handle_read()

    def doWrite(self):
        self._handle_write()
        self._manage_event_state()


class TwistedProtocolConnection(base_connection.BaseConnection):
    """A hybrid between a Pika Connection and a Twisted Protocol. Allows using
    Twisted's non-blocking connectTCP/connectSSL methods for connecting to the
    server.

    It has one caveat: TwistedProtocolConnection objects have a ready
    instance variable that's a Deferred which fires when the connection is
    ready to be used (the initial AMQP handshaking has been done). You *have*
    to wait for this Deferred to fire before requesting a channel.

    Since it's Twisted handling connection establishing it does not accept
    connect callbacks, you have to implement that within Twisted. Also remember
    that the host, port and ssl values of the connection parameters are ignored
    because, yet again, it's Twisted who manages the connection.

    """

    def __init__(self, parameters):
        self.ready = defer.Deferred()
        super(TwistedProtocolConnection, self).__init__(
            parameters=parameters,
            on_open_callback=self.connectionReady,
            on_open_error_callback=self.connectionFailed,
            on_close_callback=None,
            ioloop=IOLoopReactorAdapter(self, reactor),
            stop_ioloop_on_close=False)

    def connect(self):
        # The connection is open asynchronously by Twisted, so skip the whole
        # connect() part, except for setting the connection state
        self._set_connection_state(self.CONNECTION_INIT)

    def _adapter_connect(self):
        # Should never be called, as we override connect() and leave the
        # building of a TCP connection to Twisted, but implement anyway to keep
        # the interface
        return False

    def _adapter_disconnect(self):
        # Disconnect from the server
        self.transport.loseConnection()

    def _flush_outbound(self):
        """Override BaseConnection._flush_outbound to send all bufferred data
        the Twisted way, by writing to the transport. No need for buffering,
        Twisted handles that for us.
        """
        while self.outbound_buffer:
            self.transport.write(self.outbound_buffer.popleft())

    def channel(self, channel_number=None):
        """Create a new channel with the next available channel number or pass
        in a channel number to use. Must be non-zero if you would like to
        specify but it is recommended that you let Pika manage the channel
        numbers.

        Return a Deferred that fires with an instance of a wrapper around the
        Pika Channel class.

        :param int channel_number: The channel number to use, defaults to the
                                   next available.

        """
        d = defer.Deferred()
        base_connection.BaseConnection.channel(self, d.callback, channel_number)
        return d.addCallback(TwistedChannel)

    # IProtocol methods

    def dataReceived(self, data):
        # Pass the bytes to Pika for parsing
        self._on_data_available(data)

    def connectionLost(self, reason):
        # Let the caller know there's been an error
        d, self.ready = self.ready, None
        if d:
            d.errback(reason)

    def makeConnection(self, transport):
        self.transport = transport
        self.connectionMade()

    def connectionMade(self):
        # Tell everyone we're connected
        self._on_connected()

    # Our own methods

    def connectionReady(self, res):
        d, self.ready = self.ready, None
        if d:
            d.callback(res)

    def connectionFailed(self, connection_unused, error_message=None):
        d, self.ready = self.ready, None
        if d:
            attempts = self.params.connection_attempts
            exc = exceptions.AMQPConnectionError(attempts)
            d.errback(exc)



"""Core connection objects"""
import ast
import sys
import collections
import logging
import math
import platform
import threading
import urllib
import warnings

if sys.version_info > (3,):
    import urllib.parse as urlparse
else:
    import urlparse

from servicebus.pika import __version__
from servicebus.pika import callback
from servicebus.pika import channel
from servicebus.pika import credentials as pika_credentials
from servicebus.pika import exceptions
from servicebus.pika import frame
from servicebus.pika import heartbeat
from servicebus.pika import utils

from servicebus.pika import spec

from servicebus.pika.compat import basestring, url_unquote, dictkeys


BACKPRESSURE_WARNING = ("Pika: Write buffer exceeded warning threshold at "
                        "%i bytes and an estimated %i frames behind")
PRODUCT = "Pika Python Client Library"

LOGGER = logging.getLogger(__name__)


class Parameters(object):
    """Base connection parameters class definition

    :param str DEFAULT_HOST: 'localhost'
    :param int DEFAULT_PORT: 5672
    :param str DEFAULT_VIRTUAL_HOST: '/'
    :param str DEFAULT_USERNAME: 'guest'
    :param str DEFAULT_PASSWORD: 'guest'
    :param int DEFAULT_HEARTBEAT_INTERVAL: None
    :param int DEFAULT_CHANNEL_MAX: 0
    :param int DEFAULT_FRAME_MAX: pika.spec.FRAME_MAX_SIZE
    :param str DEFAULT_LOCALE: 'en_US'
    :param int DEFAULT_CONNECTION_ATTEMPTS: 1
    :param int|float DEFAULT_RETRY_DELAY: 2.0
    :param int|float DEFAULT_SOCKET_TIMEOUT: 0.25
    :param bool DEFAULT_SSL: False
    :param dict DEFAULT_SSL_OPTIONS: {}
    :param int DEFAULT_SSL_PORT: 5671
    :param bool DEFAULT_BACKPRESSURE_DETECTION: False

    """
    DEFAULT_BACKPRESSURE_DETECTION = False
    DEFAULT_CONNECTION_ATTEMPTS = 1
    DEFAULT_CHANNEL_MAX = 0
    DEFAULT_FRAME_MAX = spec.FRAME_MAX_SIZE
    DEFAULT_HEARTBEAT_INTERVAL = None          # accept server's proposal
    DEFAULT_HOST = 'localhost'
    DEFAULT_LOCALE = 'en_US'
    DEFAULT_PASSWORD = 'guest'
    DEFAULT_PORT = 5672
    DEFAULT_RETRY_DELAY = 2.0
    DEFAULT_SOCKET_TIMEOUT = 0.25
    DEFAULT_SSL = False
    DEFAULT_SSL_OPTIONS = {}
    DEFAULT_SSL_PORT = 5671
    DEFAULT_USERNAME = 'guest'
    DEFAULT_VIRTUAL_HOST = '/'

    def __init__(self):
        self.virtual_host = self.DEFAULT_VIRTUAL_HOST
        self.backpressure_detection = self.DEFAULT_BACKPRESSURE_DETECTION
        self.channel_max = self.DEFAULT_CHANNEL_MAX
        self.connection_attempts = self.DEFAULT_CONNECTION_ATTEMPTS
        self.credentials = self._credentials(self.DEFAULT_USERNAME,
                                             self.DEFAULT_PASSWORD)
        self.frame_max = self.DEFAULT_FRAME_MAX
        self.heartbeat = self.DEFAULT_HEARTBEAT_INTERVAL
        self.host = self.DEFAULT_HOST
        self.locale = self.DEFAULT_LOCALE
        self.port = self.DEFAULT_PORT
        self.retry_delay = self.DEFAULT_RETRY_DELAY
        self.ssl = self.DEFAULT_SSL
        self.ssl_options = self.DEFAULT_SSL_OPTIONS
        self.socket_timeout = self.DEFAULT_SOCKET_TIMEOUT

    def __repr__(self):
        """Represent the info about the instance.

        :rtype: str

        """
        return ('<%s host=%s port=%s virtual_host=%s ssl=%s>' %
                (self.__class__.__name__, self.host, self.port,
                 self.virtual_host, self.ssl))

    def _credentials(self, username, password):
        """Return a plain credentials object for the specified username and
        password.

        :param str username: The username to use
        :param str password: The password to use
        :rtype: pika_credentials.PlainCredentials

        """
        return pika_credentials.PlainCredentials(username, password)

    def _validate_backpressure(self, backpressure_detection):
        """Validate that the backpressure detection option is a bool.

        :param bool backpressure_detection: The backpressure detection value
        :rtype: bool
        :raises: TypeError

        """
        if not isinstance(backpressure_detection, bool):
            raise TypeError('backpressure detection must be a bool')
        return True

    def _validate_channel_max(self, channel_max):
        """Validate that the channel_max value is an int

        :param int channel_max: The value to validate
        :rtype: bool
        :raises: TypeError
        :raises: ValueError

        """
        if not isinstance(channel_max, int):
            raise TypeError('channel_max must be an int')
        if channel_max < 1 or channel_max > 65535:
            raise ValueError('channel_max must be <= 65535 and > 0')
        return True

    def _validate_connection_attempts(self, connection_attempts):
        """Validate that the connection_attempts value is an int

        :param int connection_attempts: The value to validate
        :rtype: bool
        :raises: TypeError
        :raises: ValueError

        """
        if not isinstance(connection_attempts, int):
            raise TypeError('connection_attempts must be an int')
        if connection_attempts < 1:
            raise ValueError('connection_attempts must be None or > 0')
        return True

    def _validate_credentials(self, credentials):
        """Validate the credentials passed in are using a valid object type.

        :param pika.credentials.Credentials credentials: Credentials to validate
        :rtype: bool
        :raises: TypeError

        """
        for credential_type in pika_credentials.VALID_TYPES:
            if isinstance(credentials, credential_type):
                return True
        raise TypeError('Credentials must be an object of type: %r' %
                        pika_credentials.VALID_TYPES)

    def _validate_frame_max(self, frame_max):
        """Validate that the frame_max value is an int and does not exceed
         the maximum frame size and is not less than the frame min size.

        :param int frame_max: The value to validate
        :rtype: bool
        :raises: TypeError
        :raises: InvalidMinimumFrameSize

        """
        if not isinstance(frame_max, int):
            raise TypeError('frame_max must be an int')
        if frame_max < spec.FRAME_MIN_SIZE:
            raise exceptions.InvalidMinimumFrameSize
        elif frame_max > spec.FRAME_MAX_SIZE:
            raise exceptions.InvalidMaximumFrameSize
        return True

    def _validate_heartbeat_interval(self, heartbeat_interval):
        """Validate that the heartbeat_interval value is an int

        :param int heartbeat_interval: The value to validate
        :rtype: bool
        :raises: TypeError
        :raises: ValueError

        """
        if not isinstance(heartbeat_interval, int):
            raise TypeError('heartbeat must be an int')
        if heartbeat_interval < 0:
            raise ValueError('heartbeat_interval must >= 0')
        return True

    def _validate_host(self, host):
        """Validate that the host value is an str

        :param str|unicode host: The value to validate
        :rtype: bool
        :raises: TypeError

        """
        if not isinstance(host, basestring):
            raise TypeError('host must be a str or unicode str')
        return True

    def _validate_locale(self, locale):
        """Validate that the locale value is an str

        :param str locale: The value to validate
        :rtype: bool
        :raises: TypeError

        """
        if not isinstance(locale, basestring):
            raise TypeError('locale must be a str')
        return True

    def _validate_port(self, port):
        """Validate that the port value is an int

        :param int port: The value to validate
        :rtype: bool
        :raises: TypeError

        """
        if not isinstance(port, int):
            raise TypeError('port must be an int')
        return True

    def _validate_retry_delay(self, retry_delay):
        """Validate that the retry_delay value is an int or float

        :param int|float retry_delay: The value to validate
        :rtype: bool
        :raises: TypeError

        """
        if not any([isinstance(retry_delay, int),
                    isinstance(retry_delay, float)]):
            raise TypeError('retry_delay must be a float or int')
        return True

    def _validate_socket_timeout(self, socket_timeout):
        """Validate that the socket_timeout value is an int or float

        :param int|float socket_timeout: The value to validate
        :rtype: bool
        :raises: TypeError

        """
        if not any([isinstance(socket_timeout, int),
                    isinstance(socket_timeout, float)]):
            raise TypeError('socket_timeout must be a float or int')
        if not socket_timeout > 0:
            raise ValueError('socket_timeout must be > 0')
        return True

    def _validate_ssl(self, ssl):
        """Validate the SSL toggle is a bool

        :param bool ssl: The SSL enabled/disabled value
        :rtype: bool
        :raises: TypeError

        """
        if not isinstance(ssl, bool):
            raise TypeError('ssl must be a bool')
        return True

    def _validate_ssl_options(self, ssl_options):
        """Validate the SSL options value is a dictionary.

        :param dict|None ssl_options: SSL Options to validate
        :rtype: bool
        :raises: TypeError

        """
        if not isinstance(ssl_options, dict) and ssl_options is not None:
            raise TypeError('ssl_options must be either None or dict')
        return True

    def _validate_virtual_host(self, virtual_host):
        """Validate that the virtual_host value is an str

        :param str virtual_host: The value to validate
        :rtype: bool
        :raises: TypeError

        """
        if not isinstance(virtual_host, basestring):
            raise TypeError('virtual_host must be a str')
        return True


class ConnectionParameters(Parameters):
    """Connection parameters object that is passed into the connection adapter
    upon construction.

    :param str host: Hostname or IP Address to connect to
    :param int port: TCP port to connect to
    :param str virtual_host: RabbitMQ virtual host to use
    :param pika.credentials.Credentials credentials: auth credentials
    :param int channel_max: Maximum number of channels to allow
    :param int frame_max: The maximum byte size for an AMQP frame
    :param int heartbeat_interval: How often to send heartbeats
    :param bool ssl: Enable SSL
    :param dict ssl_options: Arguments passed to ssl.wrap_socket as
    :param int connection_attempts: Maximum number of retry attempts
    :param int|float retry_delay: Time to wait in seconds, before the next
    :param int|float socket_timeout: Use for high latency networks
    :param str locale: Set the locale value
    :param bool backpressure_detection: Toggle backpressure detection

    """

    def __init__(self,
                 host=None,
                 port=None,
                 virtual_host=None,
                 credentials=None,
                 channel_max=None,
                 frame_max=None,
                 heartbeat_interval=None,
                 ssl=None,
                 ssl_options=None,
                 connection_attempts=None,
                 retry_delay=None,
                 socket_timeout=None,
                 locale=None,
                 backpressure_detection=None):
        """Create a new ConnectionParameters instance.

        :param str host: Hostname or IP Address to connect to
        :param int port: TCP port to connect to
        :param str virtual_host: RabbitMQ virtual host to use
        :param pika.credentials.Credentials credentials: auth credentials
        :param int channel_max: Maximum number of channels to allow
        :param int frame_max: The maximum byte size for an AMQP frame
        :param int heartbeat_interval: How often to send heartbeats.
                                  Min between this value and server's proposal
                                  will be used. Use 0 to deactivate heartbeats
                                  and None to accept server's proposal.
        :param bool ssl: Enable SSL
        :param dict ssl_options: Arguments passed to ssl.wrap_socket
        :param int connection_attempts: Maximum number of retry attempts
        :param int|float retry_delay: Time to wait in seconds, before the next
        :param int|float socket_timeout: Use for high latency networks
        :param str locale: Set the locale value
        :param bool backpressure_detection: Toggle backpressure detection

        """
        super(ConnectionParameters, self).__init__()

        # Create the default credentials object
        if not credentials:
            credentials = self._credentials(self.DEFAULT_USERNAME,
                                            self.DEFAULT_PASSWORD)

        # Assign the values
        if host and self._validate_host(host):
            self.host = host
        if port is not None and self._validate_port(port):
            self.port = port
        if virtual_host and self._validate_virtual_host(virtual_host):
            self.virtual_host = virtual_host
        if credentials and self._validate_credentials(credentials):
            self.credentials = credentials
        if channel_max is not None and self._validate_channel_max(channel_max):
            self.channel_max = channel_max
        if frame_max is not None and self._validate_frame_max(frame_max):
            self.frame_max = frame_max
        if locale and self._validate_locale(locale):
            self.locale = locale
        if (heartbeat_interval is not None and self._validate_heartbeat_interval(heartbeat_interval)):
            self.heartbeat = heartbeat_interval
        if ssl is not None and self._validate_ssl(ssl):
            self.ssl = ssl
        if ssl_options and self._validate_ssl_options(ssl_options):
            self.ssl_options = ssl_options or dict()
        if (connection_attempts is not None and self._validate_connection_attempts(connection_attempts)):
            self.connection_attempts = connection_attempts
        if retry_delay is not None and self._validate_retry_delay(retry_delay):
            self.retry_delay = retry_delay
        if (socket_timeout is not None and self._validate_socket_timeout(socket_timeout)):
            self.socket_timeout = socket_timeout
        if (backpressure_detection is not None and self._validate_backpressure(backpressure_detection)):
            self.backpressure_detection = backpressure_detection


class URLParameters(Parameters):
    """Connect to RabbitMQ via an AMQP URL in the format::

         amqp://username:password@host:port/<virtual_host>[?query-string]

    Ensure that the virtual host is URI encoded when specified. For example if
    you are using the default "/" virtual host, the value should be `%2f`.

    Valid query string values are:

        - backpressure_detection:
            Toggle backpressure detection, possible values are `t` or `f`
        - channel_max:
            Override the default maximum channel count value
        - connection_attempts:
            Specify how many times pika should try and reconnect before it gives up
        - frame_max:
            Override the default maximum frame size for communication
        - heartbeat_interval:
            Specify the number of seconds between heartbeat frames to ensure that
            the link between RabbitMQ and your application is up
        - locale:
            Override the default `en_US` locale value
        - ssl:
            Toggle SSL, possible values are `t`, `f`
        - ssl_options:
            Arguments passed to :meth:`ssl.wrap_socket`
        - retry_delay:
            The number of seconds to sleep before attempting to connect on
            connection failure.
        - socket_timeout:
            Override low level socket timeout value

    :param str url: The AMQP URL to connect to

    """

    def __init__(self, url):
        """Create a new URLParameters instance.

        :param str url: The URL value

        """
        super(URLParameters, self).__init__()
        self._process_url(url)

    def _process_url(self, url):
        """Take an AMQP URL and break it up into the various parameters.

        :param str url: The URL to parse

        """
        if url[0:4] == 'amqp':
            url = 'http' + url[4:]

        parts = urlparse.urlparse(url)

        # Handle the Protocol scheme, changing to HTTPS so urlparse doesnt barf
        if parts.scheme == 'https':
            self.ssl = True

        if self._validate_host(parts.hostname):
            self.host = parts.hostname
        if not parts.port:
            if self.ssl:
                self.port = self.DEFAULT_SSL_PORT if \
                    self.ssl else self.DEFAULT_PORT
        elif self._validate_port(parts.port):
            self.port = parts.port

        if parts.username is not None:
            self.credentials = pika_credentials.PlainCredentials(parts.username,
                                                                 parts.password)

        # Get the Virtual Host
        if len(parts.path) <= 1:
            self.virtual_host = self.DEFAULT_VIRTUAL_HOST
        else:
            path_parts = parts.path.split('/')
            virtual_host = url_unquote(path_parts[1])
            if self._validate_virtual_host(virtual_host):
                self.virtual_host = virtual_host

        # Handle query string values, validating and assigning them
        values = urlparse.parse_qs(parts.query)

        # Cast the various numeric values to the appropriate values
        for key in dictkeys(values):
            # Always reassign the first list item in query values
            values[key] = values[key].pop(0)
            if values[key].isdigit():
                values[key] = int(values[key])
            else:
                try:
                    values[key] = float(values[key])
                except ValueError:
                    pass

        if 'backpressure_detection' in values:
            if values['backpressure_detection'] == 't':
                self.backpressure_detection = True
            elif values['backpressure_detection'] == 'f':
                self.backpressure_detection = False
            else:
                raise ValueError('Invalid backpressure_detection value: %s' %
                                 values['backpressure_detection'])

        if ('channel_max' in values and self._validate_channel_max(values['channel_max'])):
            self.channel_max = values['channel_max']

        if ('connection_attempts' in values and self._validate_connection_attempts(values['connection_attempts'])):
            self.connection_attempts = values['connection_attempts']

        if ('frame_max' in values and self._validate_frame_max(values['frame_max'])):
            self.frame_max = values['frame_max']

        if ('heartbeat_interval' in values and self._validate_heartbeat_interval(values['heartbeat_interval'])):
            self.heartbeat = values['heartbeat_interval']

        if ('locale' in values and self._validate_locale(values['locale'])):
            self.locale = values['locale']

        if ('retry_delay' in values and self._validate_retry_delay(values['retry_delay'])):
            self.retry_delay = values['retry_delay']

        if ('socket_timeout' in values and self._validate_socket_timeout(values['socket_timeout'])):
            self.socket_timeout = values['socket_timeout']

        if 'ssl_options' in values:
            options = ast.literal_eval(values['ssl_options'])
            if self._validate_ssl_options(options):
                self.ssl_options = options


class Connection(object):
    """This is the core class that implements communication with RabbitMQ. This
    class should not be invoked directly but rather through the use of an
    adapter such as SelectConnection or BlockingConnection.

    :param pika.connection.Parameters parameters: Connection parameters
    :param method on_open_callback: Called when the connection is opened
    :param method on_open_error_callback: Called if the connection cant
                                   be opened
    :param method on_close_callback: Called when the connection is closed

    """
    ON_CONNECTION_BACKPRESSURE = '_on_connection_backpressure'
    ON_CONNECTION_BLOCKED = '_on_connection_blocked'
    ON_CONNECTION_CLOSED = '_on_connection_closed'
    ON_CONNECTION_ERROR = '_on_connection_error'
    ON_CONNECTION_OPEN = '_on_connection_open'
    ON_CONNECTION_UNBLOCKED = '_on_connection_unblocked'
    CONNECTION_CLOSED = 0
    CONNECTION_INIT = 1
    CONNECTION_PROTOCOL = 2
    CONNECTION_START = 3
    CONNECTION_TUNE = 4
    CONNECTION_OPEN = 5
    CONNECTION_CLOSING = 6

    def __init__(self,
                 parameters=None,
                 on_open_callback=None,
                 on_open_error_callback=None,
                 on_close_callback=None):
        """Connection initialization expects an object that has implemented the
         Parameters class and a callback function to notify when we have
         successfully connected to the AMQP Broker.

        Available Parameters classes are the ConnectionParameters class and
        URLParameters class.

        :param pika.connection.Parameters parameters: Connection parameters
        :param method on_open_callback: Called when the connection is opened
        :param method on_open_error_callback: Called if the connection cant
                                       be opened
        :param method on_close_callback: Called when the connection is closed

        """
        self._write_lock = threading.Lock()

        # Define our callback dictionary
        self.callbacks = callback.CallbackManager()

        # Add the on connection error callback
        self.callbacks.add(0, self.ON_CONNECTION_ERROR,
                           on_open_error_callback or self._on_connection_error,
                           False)

        self.heartbeat = None

        # On connection callback
        if on_open_callback:
            self.add_on_open_callback(on_open_callback)

        # On connection callback
        if on_close_callback:
            self.add_on_close_callback(on_close_callback)

        # Set our configuration options
        self.params = parameters or ConnectionParameters()

        # Initialize the connection state and connect
        self._init_connection_state()
        self.connect()

    def add_backpressure_callback(self, callback_method):
        """Call method "callback" when pika believes backpressure is being
        applied.

        :param method callback_method: The method to call

        """
        self.callbacks.add(0, self.ON_CONNECTION_BACKPRESSURE, callback_method,
                           False)

    def add_on_close_callback(self, callback_method):
        """Add a callback notification when the connection has closed. The
        callback will be passed the connection, the reply_code (int) and the
        reply_text (str), if sent by the remote server.

        :param method callback_method: Callback to call on close

        """
        self.callbacks.add(0, self.ON_CONNECTION_CLOSED, callback_method, False)

    def add_on_connection_blocked_callback(self, callback_method):
        """Add a callback to be notified when RabbitMQ has sent a
        ``Connection.Blocked`` frame indicating that RabbitMQ is low on
        resources. Publishers can use this to voluntarily suspend publishing,
        instead of relying on back pressure throttling. The callback
        will be passed the ``Connection.Blocked`` method frame.

        :param method callback_method: Callback to call on `Connection.Blocked`

        """
        self.callbacks.add(0, spec.Connection.Blocked, callback_method, False)

    def add_on_connection_unblocked_callback(self, callback_method):
        """Add a callback to be notified when RabbitMQ has sent a
        ``Connection.Unblocked`` frame letting publishers know it's ok
        to start publishing again. The callback will be passed the
        ``Connection.Unblocked`` method frame.

        :param method callback_method: Callback to call on
                                       `Connection.Unblocked`

        """
        self.callbacks.add(0, spec.Connection.Unblocked, callback_method, False)

    def add_on_open_callback(self, callback_method):
        """Add a callback notification when the connection has opened.

        :param method callback_method: Callback to call when open

        """
        self.callbacks.add(0, self.ON_CONNECTION_OPEN, callback_method, False)

    def add_on_open_error_callback(self, callback_method, remove_default=True):
        """Add a callback notification when the connection can not be opened.

        The callback method should accept the connection object that could not
        connect, and an optional error message.

        :param method callback_method: Callback to call when can't connect
        :param bool remove_default: Remove default exception raising callback

        """
        if remove_default:
            self.callbacks.remove(0, self.ON_CONNECTION_ERROR,
                                  self._on_connection_error)
        self.callbacks.add(0, self.ON_CONNECTION_ERROR, callback_method, False)

    def add_timeout(self, deadline, callback_method):
        """Adapters should override to call the callback after the
        specified number of seconds have elapsed, using a timer, or a
        thread, or similar.

        :param int deadline: The number of seconds to wait to call callback
        :param method callback_method: The callback method

        """
        raise NotImplementedError

    def channel(self, on_open_callback, channel_number=None):
        """Create a new channel with the next available channel number or pass
        in a channel number to use. Must be non-zero if you would like to
        specify but it is recommended that you let Pika manage the channel
        numbers.

        :param method on_open_callback: The callback when the channel is opened
        :param int channel_number: The channel number to use, defaults to the
                                   next available.
        :rtype: pika.channel.Channel

        """
        if not channel_number:
            channel_number = self._next_channel_number()
        self._channels[channel_number] = self._create_channel(channel_number,
                                                              on_open_callback)
        self._add_channel_callbacks(channel_number)
        self._channels[channel_number].open()
        return self._channels[channel_number]

    def close(self, reply_code=200, reply_text='Normal shutdown'):
        """Disconnect from RabbitMQ. If there are any open channels, it will
        attempt to close them prior to fully disconnecting. Channels which
        have active consumers will attempt to send a Basic.Cancel to RabbitMQ
        to cleanly stop the delivery of messages prior to closing the channel.

        :param int reply_code: The code number for the close
        :param str reply_text: The text reason for the close

        """
        if self.is_closing or self.is_closed:
            return

        if self._has_open_channels:
            self._close_channels(reply_code, reply_text)

        # Set our connection state
        self._set_connection_state(self.CONNECTION_CLOSING)
        LOGGER.debug("Closing connection (%s): %s", reply_code, reply_text)
        self.closing = reply_code, reply_text

        if not self._has_open_channels:
            # if there are open channels then _on_close_ready will finally be
            # called in _on_channel_cleanup once all channels have been closed
            self._on_close_ready()

    def connect(self):
        """Invoke if trying to reconnect to a RabbitMQ server. Constructing the
        Connection object should connect on its own.

        """
        self._set_connection_state(self.CONNECTION_INIT)
        error = self._adapter_connect()
        if not error:
            return self._on_connected()
        self.remaining_connection_attempts -= 1
        LOGGER.warning('Could not connect, %i attempts left',
                       self.remaining_connection_attempts)
        if self.remaining_connection_attempts:
            LOGGER.debug('Retrying in %i seconds', self.params.retry_delay)
            self.add_timeout(self.params.retry_delay, self.connect)
        else:
            self.callbacks.process(0, self.ON_CONNECTION_ERROR, self, self,
                                   error)
            self.remaining_connection_attempts = self.params.connection_attempts
            self._set_connection_state(self.CONNECTION_CLOSED)

    def remove_timeout(self, callback_method):
        """Adapters should override to call the callback after the
        specified number of seconds have elapsed, using a timer, or a
        thread, or similar.

        :param method callback_method: The callback to remove a timeout for

        """
        raise NotImplementedError

    def set_backpressure_multiplier(self, value=10):
        """Alter the backpressure multiplier value. We set this to 10 by default.
        This value is used to raise warnings and trigger the backpressure
        callback.

        :param int value: The multiplier value to set

        """
        self._backpressure = value

    #
    # Connections state properties
    #

    @property
    def is_closed(self):
        """
        Returns a boolean reporting the current connection state.
        """
        return self.connection_state == self.CONNECTION_CLOSED

    @property
    def is_closing(self):
        """
        Returns a boolean reporting the current connection state.
        """
        return self.connection_state == self.CONNECTION_CLOSING

    @property
    def is_open(self):
        """
        Returns a boolean reporting the current connection state.
        """
        return self.connection_state == self.CONNECTION_OPEN

    #
    # Properties that reflect server capabilities for the current connection
    #

    @property
    def basic_nack(self):
        """Specifies if the server supports basic.nack on the active connection.

        :rtype: bool

        """
        return self.server_capabilities.get('basic.nack', False)

    @property
    def consumer_cancel_notify(self):
        """Specifies if the server supports consumer cancel notification on the
        active connection.

        :rtype: bool

        """
        return self.server_capabilities.get('consumer_cancel_notify', False)

    @property
    def exchange_exchange_bindings(self):
        """Specifies if the active connection supports exchange to exchange
        bindings.

        :rtype: bool

        """
        return self.server_capabilities.get('exchange_exchange_bindings', False)

    @property
    def publisher_confirms(self):
        """Specifies if the active connection can use publisher confirmations.

        :rtype: bool

        """
        return self.server_capabilities.get('publisher_confirms', False)

    #
    # Internal methods for managing the communication process
    #

    def _adapter_connect(self):
        """Subclasses should override to set up the outbound socket connection.

        :raises: NotImplementedError

        """
        raise NotImplementedError

    def _adapter_disconnect(self):
        """Subclasses should override this to cause the underlying transport
        (socket) to close.

        :raises: NotImplementedError

        """
        raise NotImplementedError

    def _add_channel_callbacks(self, channel_number):
        """Add the appropriate callbacks for the specified channel number.

        :param int channel_number: The channel number for the callbacks

        """
        # This permits us to garbage-collect our reference to the channel
        # regardless of whether it was closed by client or broker, and do so
        # after all channel-close callbacks.
        self._channels[channel_number]._add_on_cleanup_callback(
            self._on_channel_cleanup)

    def _add_connection_start_callback(self):
        """Add a callback for when a Connection.Start frame is received from
        the broker.

        """
        self.callbacks.add(0, spec.Connection.Start, self._on_connection_start)

    def _add_connection_tune_callback(self):
        """Add a callback for when a Connection.Tune frame is received."""
        self.callbacks.add(0, spec.Connection.Tune, self._on_connection_tune)

    def _append_frame_buffer(self, value):
        """Append the bytes to the frame buffer.

        :param str value: The bytes to append to the frame buffer

        """
        self._frame_buffer += value

    @property
    def _buffer_size(self):
        """Return the suggested buffer size from the connection state/tune or
        the default if that is None.

        :rtype: int

        """
        return self.params.frame_max or spec.FRAME_MAX_SIZE

    def _check_for_protocol_mismatch(self, value):
        """Invoked when starting a connection to make sure it's a supported
        protocol.

        :param pika.frame.Method value: The frame to check
        :raises: ProtocolVersionMismatch

        """
        if (value.method.version_major, value.method.version_minor) != spec.PROTOCOL_VERSION[0:2]:
            raise exceptions.ProtocolVersionMismatch(frame.ProtocolHeader(),
                                                     value)

    @property
    def _client_properties(self):
        """Return the client properties dictionary.

        :rtype: dict

        """
        return {
            'product': PRODUCT,
            'platform': 'Python %s' % platform.python_version(),
            'capabilities': {
                'authentication_failure_close': True,
                'basic.nack': True,
                'connection.blocked': True,
                'consumer_cancel_notify': True,
                'publisher_confirms': True
            },
            'information': 'See http://pika.rtfd.org',
            'version': __version__
        }

    def _close_channels(self, reply_code, reply_text):
        """Close the open channels with the specified reply_code and reply_text.

        :param int reply_code: The code for why the channels are being closed
        :param str reply_text: The text reason for why the channels are closing

        """
        if self.is_open:
            for channel_number in dictkeys(self._channels):
                if self._channels[channel_number].is_open:
                    self._channels[channel_number].close(reply_code, reply_text)
                else:
                    del self._channels[channel_number]
                    # Force any lingering callbacks to be removed
                    # moved inside else block since channel's _cleanup removes
                    # callbacks
                    self.callbacks.cleanup(channel_number)
        else:
            self._channels = dict()

    def _combine(self, a, b):
        """Pass in two values, if a is 0, return b otherwise if b is 0,
        return a. If neither case matches return the smallest value.

        :param int a: The first value
        :param int b: The second value
        :rtype: int

        """
        return min(a, b) or (a or b)

    def _connect(self):
        """Attempt to connect to RabbitMQ

        :rtype: bool

        """
        warnings.warn('This method is deprecated, use Connection.connect',
                      DeprecationWarning)

    def _create_channel(self, channel_number, on_open_callback):
        """Create a new channel using the specified channel number and calling
        back the method specified by on_open_callback

        :param int channel_number: The channel number to use
        :param method on_open_callback: The callback when the channel is opened

        """
        LOGGER.debug('Creating channel %s', channel_number)
        return channel.Channel(self, channel_number, on_open_callback)

    def _create_heartbeat_checker(self):
        """Create a heartbeat checker instance if there is a heartbeat interval
        set.

        :rtype: pika.heartbeat.Heartbeat

        """
        if self.params.heartbeat is not None and self.params.heartbeat > 0:
            LOGGER.debug('Creating a HeartbeatChecker: %r',
                         self.params.heartbeat)
            return heartbeat.HeartbeatChecker(self, self.params.heartbeat)

    def _remove_heartbeat(self):
        """Stop the heartbeat checker if it exists

        """
        if self.heartbeat:
            self.heartbeat.stop()
            self.heartbeat = None

    def _deliver_frame_to_channel(self, value):
        """Deliver the frame to the channel specified in the frame.

        :param pika.frame.Method value: The frame to deliver

        """
        if value.channel_number not in self._channels:
            if self._is_basic_deliver_frame(value):
                self._reject_out_of_band_delivery(value.channel_number,
                                                  value.method.delivery_tag)
            else:
                LOGGER.warning("Received %r for non-existing channel %i", value,
                               value.channel_number)
            return
        return self._channels[value.channel_number]._handle_content_frame(value)

    def _detect_backpressure(self):
        """Attempt to calculate if TCP backpressure is being applied due to
        our outbound buffer being larger than the average frame size over
        a window of frames.

        """
        avg_frame_size = self.bytes_sent / self.frames_sent
        buffer_size = sum([len(frame) for frame in self.outbound_buffer])
        if buffer_size > (avg_frame_size * self._backpressure):
            LOGGER.warning(BACKPRESSURE_WARNING, buffer_size,
                           int(buffer_size / avg_frame_size))
            self.callbacks.process(0, self.ON_CONNECTION_BACKPRESSURE, self)

    def _ensure_closed(self):
        """If the connection is not closed, close it."""
        if self.is_open:
            self.close()

    def _flush_outbound(self):
        """Adapters should override to flush the contents of outbound_buffer
        out along the socket.

        :raises: NotImplementedError

        """
        raise NotImplementedError

    def _get_body_frame_max_length(self):
        """Calculate the maximum amount of bytes that can be in a body frame.

        :rtype: int

        """
        return (
            self.params.frame_max - spec.FRAME_HEADER_SIZE - spec.FRAME_END_SIZE
        )

    def _get_credentials(self, method_frame):
        """Get credentials for authentication.

        :param pika.frame.MethodFrame method_frame: The Connection.Start frame
        :rtype: tuple(str, str)

        """
        (auth_type,
         response) = self.params.credentials.response_for(method_frame.method)
        if not auth_type:
            raise exceptions.AuthenticationError(self.params.credentials.TYPE)
        self.params.credentials.erase_credentials()
        return auth_type, response

    @property
    def _has_open_channels(self):
        """Returns true if channels are open.

        :rtype: bool

        """
        return any([self._channels[num].is_open
                    for num in dictkeys(self._channels)])

    def _has_pending_callbacks(self, value):
        """Return true if there are any callbacks pending for the specified
        frame.

        :param pika.frame.Method value: The frame to check
        :rtype: bool

        """
        return self.callbacks.pending(value.channel_number, value.method)

    def _init_connection_state(self):
        """Initialize or reset all of the internal state variables for a given
        connection. On disconnect or reconnect all of the state needs to
        be wiped.

        """
        # Connection state
        self._set_connection_state(self.CONNECTION_CLOSED)

        # Negotiated server properties
        self.server_properties = None

        # Outbound buffer for buffering writes until we're able to send them
        self.outbound_buffer = collections.deque([])

        # Inbound buffer for decoding frames
        self._frame_buffer = bytes()

        # Dict of open channels
        self._channels = dict()

        # Remaining connection attempts
        self.remaining_connection_attempts = self.params.connection_attempts

        # Data used for Heartbeat checking and back-pressure detection
        self.bytes_sent = 0
        self.bytes_received = 0
        self.frames_sent = 0
        self.frames_received = 0
        self.heartbeat = None

        # Default back-pressure multiplier value
        self._backpressure = 10

        # When closing, hold reason why
        self.closing = 0, 'Not specified'

        # Our starting point once connected, first frame received
        self._add_connection_start_callback()

    def _is_basic_deliver_frame(self, frame_value):
        """Returns true if the frame is a Basic.Deliver

        :param pika.frame.Method frame_value: The frame to check
        :rtype: bool

        """
        return isinstance(frame_value, spec.Basic.Deliver)

    def _is_connection_close_frame(self, value):
        """Returns true if the frame is a Connection.Close frame.

        :param pika.frame.Method value: The frame to check
        :rtype: bool

        """
        if not value:
            return False
        return isinstance(value.method, spec.Connection.Close)

    def _is_method_frame(self, value):
        """Returns true if the frame is a method frame.

        :param pika.frame.Frame value: The frame to evaluate
        :rtype: bool

        """
        return isinstance(value, frame.Method)

    def _is_protocol_header_frame(self, value):
        """Returns True if it's a protocol header frame.

        :rtype: bool

        """
        return isinstance(value, frame.ProtocolHeader)

    def _next_channel_number(self):
        """Return the next available channel number or raise an exception.

        :rtype: int

        """
        limit = self.params.channel_max or channel.MAX_CHANNELS
        if len(self._channels) == limit:
            raise exceptions.NoFreeChannels()

        ckeys = set(self._channels.keys())
        if not ckeys:
            return 1
        return [x + 1 for x in sorted(ckeys) if x + 1 not in ckeys][0]

    def _on_channel_cleanup(self, channel):
        """Remove the channel from the dict of channels when Channel.CloseOk is
        sent. If connection is closing and no more channels remain, proceed to
        `_on_close_ready`.

        :param pika.channel.Channel channel: channel instance

        """
        try:
            del self._channels[channel.channel_number]
            LOGGER.debug('Removed channel %s', channel.channel_number)
        except KeyError:
            LOGGER.error('Channel %r not in channels',
                         channel.channel_number)
        if self.is_closing and not self._has_open_channels:
            self._on_close_ready()

    def _on_close_ready(self):
        """Called when the Connection is in a state that it can close after
        a close has been requested. This happens, for example, when all of the
        channels are closed that were open when the close request was made.

        """
        if self.is_closed:
            LOGGER.warning('Invoked while already closed')
            return
        self._send_connection_close(self.closing[0], self.closing[1])

    def _on_connected(self):
        """Invoked when the socket is connected and it's time to start speaking
        AMQP with the broker.

        """
        self._set_connection_state(self.CONNECTION_PROTOCOL)

        # Start the communication with the RabbitMQ Broker
        self._send_frame(frame.ProtocolHeader())

    def _on_connection_closed(self, method_frame, from_adapter=False):
        """Called when the connection is closed remotely. The from_adapter value
        will be true if the connection adapter has been disconnected from
        the broker and the method was invoked directly instead of by receiving
        a Connection.Close frame.

        :param pika.frame.Method: The Connection.Close frame
        :param bool from_adapter: Called by the connection adapter

        """
        if method_frame and self._is_connection_close_frame(method_frame):
            self.closing = (method_frame.method.reply_code,
                            method_frame.method.reply_text)

        # Save the codes because self.closing gets reset by _adapter_disconnect
        reply_code, reply_text = self.closing

        # Stop the heartbeat checker if it exists
        self._remove_heartbeat()

        # If this did not come from the connection adapter, close the socket
        if not from_adapter:
            self._adapter_disconnect()

        # Invoke a method frame neutral close
        self._on_disconnect(reply_code, reply_text)

    def _on_connection_error(self, connection_unused, error_message=None):
        """Default behavior when the connecting connection can not connect.

        :raises: exceptions.AMQPConnectionError

        """
        raise exceptions.AMQPConnectionError(error_message or
                                             self.params.connection_attempts)

    def _on_connection_open(self, method_frame):
        """
        This is called once we have tuned the connection with the server and
        called the Connection.Open on the server and it has replied with
        Connection.Ok.
        """
        self.known_hosts = method_frame.method.known_hosts

        # Add a callback handler for the Broker telling us to disconnect
        self.callbacks.add(0, spec.Connection.Close, self._on_connection_closed)

        # We're now connected at the AMQP level
        self._set_connection_state(self.CONNECTION_OPEN)

        # Call our initial callback that we're open
        self.callbacks.process(0, self.ON_CONNECTION_OPEN, self, self)

    def _on_connection_start(self, method_frame):
        """This is called as a callback once we have received a Connection.Start
        from the server.

        :param pika.frame.Method method_frame: The frame received
        :raises: UnexpectedFrameError

        """
        self._set_connection_state(self.CONNECTION_START)
        if self._is_protocol_header_frame(method_frame):
            raise exceptions.UnexpectedFrameError
        self._check_for_protocol_mismatch(method_frame)
        self._set_server_information(method_frame)
        self._add_connection_tune_callback()
        self._send_connection_start_ok(*self._get_credentials(method_frame))

    def _on_connection_tune(self, method_frame):
        """Once the Broker sends back a Connection.Tune, we will set our tuning
        variables that have been returned to us and kick off the Heartbeat
        monitor if required, send our TuneOk and then the Connection. Open rpc
        call on channel 0.

        :param pika.frame.Method method_frame: The frame received

        """
        self._set_connection_state(self.CONNECTION_TUNE)

        # Get our max channels, frames and heartbeat interval
        self.params.channel_max = self._combine(self.params.channel_max,
                                                method_frame.method.channel_max)
        self.params.frame_max = self._combine(self.params.frame_max,
                                              method_frame.method.frame_max)
        if self.params.heartbeat is None:
            self.params.heartbeat = method_frame.method.heartbeat
        elif self.params.heartbeat != 0:
            self.params.heartbeat = self._combine(self.params.heartbeat,
                                                  method_frame.method.heartbeat)

        # Calculate the maximum pieces for body frames
        self._body_max_length = self._get_body_frame_max_length()

        # Create a new heartbeat checker if needed
        self.heartbeat = self._create_heartbeat_checker()

        # Send the TuneOk response with what we've agreed upon
        self._send_connection_tune_ok()

        # Send the Connection.Open RPC call for the vhost
        self._send_connection_open()

    def _on_data_available(self, data_in):
        """This is called by our Adapter, passing in the data from the socket.
        As long as we have buffer try and map out frame data.

        :param str data_in: The data that is available to read

        """
        self._append_frame_buffer(data_in)
        while self._frame_buffer:
            consumed_count, frame_value = self._read_frame()
            if not frame_value:
                return
            self._trim_frame_buffer(consumed_count)
            self._process_frame(frame_value)

    def _on_disconnect(self, reply_code, reply_text):
        """Invoke passing in the reply_code and reply_text from internal
        methods to the adapter. Called from on_connection_closed and Heartbeat
        timeouts.

        :param str reply_code: The numeric close code
        :param str reply_text: The text close reason

        """
        LOGGER.debug('Disconnected from RabbitMQ at %s:%i (%s): %s',
                     self.params.host, self.params.port, reply_code,
                     reply_text)
        self._set_connection_state(self.CONNECTION_CLOSED)
        for channel in dictkeys(self._channels):
            if channel not in self._channels:
                continue
            method_frame = frame.Method(channel, spec.Channel.Close(reply_code,
                                                                    reply_text))
            self._channels[channel]._on_close(method_frame)
        self._process_connection_closed_callbacks(reply_code, reply_text)
        self._remove_connection_callbacks()

    def _process_callbacks(self, frame_value):
        """Process the callbacks for the frame if the frame is a method frame
        and if it has any callbacks pending.

        :param pika.frame.Method frame_value: The frame to process
        :rtype: bool

        """
        if (self._is_method_frame(frame_value) and self._has_pending_callbacks(frame_value)):
            self.callbacks.process(frame_value.channel_number,  # Prefix
                                   frame_value.method,  # Key
                                   self,  # Caller
                                   frame_value)  # Args
            return True
        return False

    def _process_connection_closed_callbacks(self, reason_code, reason_text):
        """Process any callbacks that should be called when the connection is
        closed.

        :param str reason_code: The numeric code from RabbitMQ for the close
        :param str reason_text: The text reason fro closing

        """
        self.callbacks.process(0, self.ON_CONNECTION_CLOSED, self, self,
                               reason_code, reason_text)

    def _process_frame(self, frame_value):
        """Process an inbound frame from the socket.

        :param frame_value: The frame to process
        :type frame_value: pika.frame.Frame | pika.frame.Method

        """
        # Will receive a frame type of -1 if protocol version mismatch
        if frame_value.frame_type < 0:
            return

        # Keep track of how many frames have been read
        self.frames_received += 1

        # Process any callbacks, if True, exit method
        if self._process_callbacks(frame_value):
            return

        # If a heartbeat is received, update the checker
        if isinstance(frame_value, frame.Heartbeat):
            if self.heartbeat:
                self.heartbeat.received()
            else:
                LOGGER.warning('Received heartbeat frame without a heartbeat '
                               'checker')

        # If the frame has a channel number beyond the base channel, deliver it
        elif frame_value.channel_number > 0:
            self._deliver_frame_to_channel(frame_value)

    def _read_frame(self):
        """Try and read from the frame buffer and decode a frame.

        :rtype tuple: (int, pika.frame.Frame)

        """
        return frame.decode_frame(self._frame_buffer)

    def _reject_out_of_band_delivery(self, channel_number, delivery_tag):
        """Reject a delivery on the specified channel number and delivery tag
        because said channel no longer exists.

        :param int channel_number: The channel number
        :param int delivery_tag: The delivery tag

        """
        LOGGER.warning('Rejected out-of-band delivery on channel %i (%s)',
                       channel_number, delivery_tag)
        self._send_method(channel_number, spec.Basic.Reject(delivery_tag))

    def _remove_callback(self, channel_number, method_frame):
        """Remove the specified method_frame callback if it is set for the
        specified channel number.

        :param int channel_number: The channel number to remove the callback on
        :param pika.object.Method: The method frame for the callback

        """
        self.callbacks.remove(str(channel_number), method_frame)

    def _remove_callbacks(self, channel_number, method_frames):
        """Remove the callbacks for the specified channel number and list of
        method frames.

        :param int channel_number: The channel number to remove the callback on
        :param list method_frames: The method frames for the callback

        """
        for method_frame in method_frames:
            self._remove_callback(channel_number, method_frame)

    def _remove_connection_callbacks(self):
        """Remove all callbacks for the connection"""
        self._remove_callbacks(0, [spec.Connection.Close, spec.Connection.Start,
                                   spec.Connection.Open])

    def _rpc(self, channel_number, method_frame,
             callback_method=None,
             acceptable_replies=None):
        """Make an RPC call for the given callback, channel number and method.
        acceptable_replies lists out what responses we'll process from the
        server with the specified callback.

        :param int channel_number: The channel number for the RPC call
        :param pika.object.Method method_frame: The method frame to call
        :param method callback_method: The callback for the RPC response
        :param list acceptable_replies: The replies this RPC call expects

        """
        # Validate that acceptable_replies is a list or None
        if acceptable_replies and not isinstance(acceptable_replies, list):
            raise TypeError('acceptable_replies should be list or None')

        # Validate the callback is callable
        if callback_method:
            if not utils.is_callable(callback_method):
                raise TypeError('callback should be None, function or method.')

            for reply in acceptable_replies:
                self.callbacks.add(channel_number, reply, callback_method)

        # Send the rpc call to RabbitMQ
        self._send_method(channel_number, method_frame)

    def _send_connection_close(self, reply_code, reply_text):
        """Send a Connection.Close method frame.

        :param int reply_code: The reason for the close
        :param str reply_text: The text reason for the close

        """
        self._rpc(0, spec.Connection.Close(reply_code, reply_text, 0, 0),
                  self._on_connection_closed, [spec.Connection.CloseOk])

    def _send_connection_open(self):
        """Send a Connection.Open frame"""
        self._rpc(0, spec.Connection.Open(self.params.virtual_host,
                                          insist=True),
                  self._on_connection_open, [spec.Connection.OpenOk])

    def _send_connection_start_ok(self, authentication_type, response):
        """Send a Connection.StartOk frame

        :param str authentication_type: The auth type value
        :param str response: The encoded value to send

        """
        self._send_method(0,
                          spec.Connection.StartOk(self._client_properties,
                                                  authentication_type, response,
                                                  self.params.locale))

    def _send_connection_tune_ok(self):
        """Send a Connection.TuneOk frame"""
        self._send_method(0, spec.Connection.TuneOk(self.params.channel_max,
                                                    self.params.frame_max,
                                                    self.params.heartbeat))

    def _send_frame(self, frame_value):
        """This appends the fully generated frame to send to the broker to the
        output buffer which will be then sent via the connection adapter.

        :param frame_value: The frame to write
        :type frame_value:  pika.frame.Frame|pika.frame.ProtocolHeader
        :raises: exceptions.ConnectionClosed

        """
        if self.is_closed:
            LOGGER.critical('Attempted to send frame when closed')
            raise exceptions.ConnectionClosed

        marshaled_frame = frame_value.marshal()
        self.bytes_sent += len(marshaled_frame)
        self.frames_sent += 1
        self.outbound_buffer.append(marshaled_frame)
        self._flush_outbound()
        if self.params.backpressure_detection:
            self._detect_backpressure()

    def _send_method(self, channel_number, method_frame, content=None):
        """Constructs a RPC method frame and then sends it to the broker.

        :param int channel_number: The channel number for the frame
        :param pika.object.Method method_frame: The method frame to send
        :param tuple content: If set, is a content frame, is tuple of
                              properties and body.

        """
        if not content:
            with self._write_lock:
                self._send_frame(frame.Method(channel_number, method_frame))
                return
        self._send_message(channel_number, method_frame, content)

    def _send_message(self, channel_number, method_frame, content=None):
        """Send the message directly, bypassing the single _send_frame
        invocation by directly appending to the output buffer and flushing
        within a lock.

        :param int channel_number: The channel number for the frame
        :param pika.object.Method method_frame: The method frame to send
        :param tuple content: If set, is a content frame, is tuple of
                              properties and body.

        """
        length = len(content[1])
        write_buffer = [frame.Method(channel_number, method_frame).marshal(),
                        frame.Header(channel_number, length,
                                     content[0]).marshal()]
        if content[1]:
            chunks = int(math.ceil(float(length) / self._body_max_length))
            for chunk in range(0, chunks):
                s = chunk * self._body_max_length
                e = s + self._body_max_length
                if e > length:
                    e = length
                write_buffer.append(frame.Body(channel_number,
                                               content[1][s:e]).marshal())

        with self._write_lock:
            self.outbound_buffer += write_buffer
            self.frames_sent += len(write_buffer)
            self._flush_outbound()
            if self.params.backpressure_detection:
                self._detect_backpressure()

    def _set_connection_state(self, connection_state):
        """Set the connection state.

        :param int connection_state: The connection state to set

        """
        self.connection_state = connection_state

    def _set_server_information(self, method_frame):
        """Set the server properties and capabilities

        :param spec.connection.Start method_frame: The Connection.Start frame

        """
        self.server_properties = method_frame.method.server_properties
        self.server_capabilities = self.server_properties.get('capabilities',
                                                              dict())
        if hasattr(self.server_properties, 'capabilities'):
            del self.server_properties['capabilities']

    def _trim_frame_buffer(self, byte_count):
        """Trim the leading N bytes off the frame buffer and increment the
        counter that keeps track of how many bytes have been read/used from the
        socket.

        :param int byte_count: The number of bytes consumed

        """
        self._frame_buffer = self._frame_buffer[byte_count:]
        self.bytes_received += byte_count



__version__ = '0.10.0'

import logging
try:
    # not available in python 2.6
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):

        def emit(self, record):
            pass

# Add NullHandler to prevent logging warnings
logging.getLogger(__name__).addHandler(NullHandler())

from servicebus.pika.connection import ConnectionParameters
from servicebus.pika.connection import URLParameters
from servicebus.pika.credentials import PlainCredentials
from servicebus.pika.spec import BasicProperties

from servicebus.pika.adapters import BaseConnection
from servicebus.pika.adapters import BlockingConnection
from servicebus.pika.adapters import SelectConnection
from servicebus.pika.adapters import TornadoConnection
from servicebus.pika.adapters import TwistedConnection
from servicebus.pika.adapters import LibevConnection



"""A connection adapter that tries to use the best polling method for the
platform pika is running on.

"""
import os
import logging
import socket
import select
import errno
import time
from collections import defaultdict
import threading

from servicebus.pika import compat
from servicebus.pika.compat import dictkeys

from servicebus.pika.adapters.base_connection import BaseConnection

LOGGER = logging.getLogger(__name__)

# One of select, epoll, kqueue or poll
SELECT_TYPE = None

# Use epoll's constants to keep life easy
READ = 0x0001
WRITE = 0x0004
ERROR = 0x0008


if compat.PY2:
    _SELECT_ERROR = select.error
else:
    # select.error was deprecated and replaced by OSError in python 3.3
    _SELECT_ERROR = OSError


def _get_select_errno(error):
    if compat.PY2:
        assert isinstance(error, select.error), repr(error)
        return error.args[0]
    else:
        assert isinstance(error, OSError), repr(error)
        return error.errno


class SelectConnection(BaseConnection):
    """An asynchronous connection adapter that attempts to use the fastest
    event loop adapter for the given platform.

    """

    def __init__(self,
                 parameters=None,
                 on_open_callback=None,
                 on_open_error_callback=None,
                 on_close_callback=None,
                 stop_ioloop_on_close=True,
                 custom_ioloop=None):
        """Create a new instance of the Connection object.

        :param pika.connection.Parameters parameters: Connection parameters
        :param method on_open_callback: Method to call on connection open
        :param on_open_error_callback: Method to call if the connection cant
                                       be opened
        :type on_open_error_callback: method
        :param method on_close_callback: Method to call on connection close
        :param bool stop_ioloop_on_close: Call ioloop.stop() if disconnected
        :param custom_ioloop: Override using the global IOLoop in Tornado
        :raises: RuntimeError

        """
        ioloop = custom_ioloop or IOLoop()
        super(SelectConnection, self).__init__(parameters, on_open_callback,
                                               on_open_error_callback,
                                               on_close_callback, ioloop,
                                               stop_ioloop_on_close)

    def _adapter_connect(self):
        """Connect to the RabbitMQ broker, returning True on success, False
        on failure.

        :rtype: bool

        """
        error = super(SelectConnection, self)._adapter_connect()
        if not error:
            self.ioloop.add_handler(self.socket.fileno(), self._handle_events,
                                    self.event_state)
        return error

    def _adapter_disconnect(self):
        """Disconnect from the RabbitMQ broker"""
        if self.socket:
            self.ioloop.remove_handler(self.socket.fileno())
        super(SelectConnection, self)._adapter_disconnect()


class IOLoop(object):
    """Singlton wrapper that decides which type of poller to use, creates an
    instance of it in start_poller and keeps the invoking application in a
    blocking state by calling the pollers start method. Poller should keep
    looping until IOLoop.instance().stop() is called or there is a socket
    error.

    Passes through all operations to the loaded poller object.

    """

    def __init__(self):
        self._poller = self._get_poller()

    def __getattr__(self, attr):
        return getattr(self._poller, attr)

    def _get_poller(self):
        """Determine the best poller to use for this enviroment."""

        poller = None

        if hasattr(select, 'epoll'):
            if not SELECT_TYPE or SELECT_TYPE == 'epoll':
                LOGGER.debug('Using EPollPoller')
                poller = EPollPoller()

        if not poller and hasattr(select, 'kqueue'):
            if not SELECT_TYPE or SELECT_TYPE == 'kqueue':
                LOGGER.debug('Using KQueuePoller')
                poller = KQueuePoller()

        if (not poller and hasattr(select, 'poll') and
                hasattr(select.poll(), 'modify')):  # pylint: disable=E1101
            if not SELECT_TYPE or SELECT_TYPE == 'poll':
                LOGGER.debug('Using PollPoller')
                poller = PollPoller()

        if not poller:
            LOGGER.debug('Using SelectPoller')
            poller = SelectPoller()

        return poller


class SelectPoller(object):
    """Default behavior is to use Select since it's the widest supported and has
    all of the methods we need for child classes as well. One should only need
    to override the update_handler and start methods for additional types.

    """
    # Drop out of the poll loop every POLL_TIMEOUT secs as a worst case, this
    # is only a backstop value. We will run timeouts when they are scheduled.
    POLL_TIMEOUT = 5
    # if the poller uses MS specify 1000
    POLL_TIMEOUT_MULT = 1

    def __init__(self):
        """Create an instance of the SelectPoller

        """
        # fd-to-handler function mappings
        self._fd_handlers = dict()

        # event-to-fdset mappings
        self._fd_events = {READ: set(), WRITE: set(), ERROR: set()}

        self._stopping = False
        self._timeouts = {}
        self._next_timeout = None
        self._processing_fd_event_map = {}

        # Mutex for controlling critical sections where ioloop-interrupt sockets
        # are created, used, and destroyed. Needed in case `stop()` is called
        # from a thread.
        self._mutex = threading.Lock()

        # ioloop-interrupt socket pair; initialized in start()
        self._r_interrupt = None
        self._w_interrupt = None

    def get_interrupt_pair(self):
        """ Use a socketpair to be able to interrupt the ioloop if called
            from another thread. Socketpair() is not supported on some OS (Win)
            so use a pair of simple UDP sockets instead. The sockets will be
            closed and garbage collected by python when the ioloop itself is.
        """
        try:
            read_sock, write_sock = socket.socketpair()

        except AttributeError:
            LOGGER.debug("Using custom socketpair for interrupt")
            read_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            read_sock.bind(('localhost', 0))
            write_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            write_sock.connect(read_sock.getsockname())

        read_sock.setblocking(0)
        write_sock.setblocking(0)
        return read_sock, write_sock

    def read_interrupt(self, interrupt_sock,
                       events, write_only):  # pylint: disable=W0613
        """ Read the interrupt byte(s). We ignore the event mask and write_only
        flag as we can ony get here if there's data to be read on our fd.

        :param int interrupt_sock: The file descriptor to read from
        :param int events: (unused) The events generated for this fd
        :param bool write_only: (unused) True if poll was called to trigger a
            write
        """
        try:
            os.read(interrupt_sock, 512)
        except OSError as err:
            if err.errno != errno.EAGAIN:
                raise

    def add_timeout(self, deadline, callback_method):
        """Add the callback_method to the IOLoop timer to fire after deadline
        seconds. Returns a handle to the timeout. Do not confuse with
        Tornado's timeout where you pass in the time you want to have your
        callback called. Only pass in the seconds until it's to be called.

        :param int deadline: The number of seconds to wait to call callback
        :param method callback_method: The callback method
        :rtype: str

        """
        timeout_at = time.time() + deadline
        value = {'deadline': timeout_at, 'callback': callback_method}
        timeout_id = hash(frozenset(value.items()))
        self._timeouts[timeout_id] = value

        if not self._next_timeout or timeout_at < self._next_timeout:
            self._next_timeout = timeout_at

        return timeout_id

    def remove_timeout(self, timeout_id):
        """Remove a timeout if it's still in the timeout stack

        :param str timeout_id: The timeout id to remove

        """
        try:
            timeout = self._timeouts.pop(timeout_id)
            if timeout['deadline'] == self._next_timeout:
                self._next_timeout = None
        except KeyError:
            pass

    def get_next_deadline(self):
        """Get the interval to the next timeout event, or a default interval
        """
        if self._next_timeout:
            timeout = max((self._next_timeout - time.time(), 0))

        elif self._timeouts:
            deadlines = [t['deadline'] for t in self._timeouts.values()]
            self._next_timeout = min(deadlines)
            timeout = max((self._next_timeout - time.time(), 0))

        else:
            timeout = SelectPoller.POLL_TIMEOUT

        timeout = min((timeout, SelectPoller.POLL_TIMEOUT))
        return timeout * SelectPoller.POLL_TIMEOUT_MULT

    def process_timeouts(self):
        """Process the self._timeouts event stack"""

        now = time.time()
        # Run the timeouts in order of deadlines. Although this shouldn't
        # be strictly necessary it preserves old behaviour when timeouts
        # were only run periodically.
        to_run = sorted([(k, timer) for (k, timer) in self._timeouts.items()
                         if timer['deadline'] <= now],
                        key=lambda item: item[1]['deadline'])

        for k, timer in to_run:
            if k not in self._timeouts:
                # Previous invocation(s) should have deleted the timer.
                continue
            try:
                timer['callback']()
            finally:
                # Don't do 'del self._timeout[k]' as the key might
                # have been deleted just now.
                if self._timeouts.pop(k, None) is not None:
                    self._next_timeout = None

    def add_handler(self, fileno, handler, events):
        """Add a new fileno to the set to be monitored

       :param int fileno: The file descriptor
       :param method handler: What is called when an event happens
       :param int events: The event mask

       """
        self._fd_handlers[fileno] = handler
        self.update_handler(fileno, events)

    def update_handler(self, fileno, events):
        """Set the events to the current events

        :param int fileno: The file descriptor
        :param int events: The event mask

        """

        for ev in (READ, WRITE, ERROR):
            if events & ev:
                self._fd_events[ev].add(fileno)
            else:
                self._fd_events[ev].discard(fileno)

    def remove_handler(self, fileno):
        """Remove a file descriptor from the set

        :param int fileno: The file descriptor

        """
        try:
            del self._processing_fd_event_map[fileno]
        except KeyError:
            pass

        self.update_handler(fileno, 0)
        del self._fd_handlers[fileno]

    def start(self):
        """Start the main poller loop. It will loop here until self._stopping"""

        LOGGER.debug('Starting IOLoop')
        self._stopping = False

        with self._mutex:
            # Watch out for reentry
            if self._r_interrupt is None:
                # Create ioloop-interrupt socket pair and register read handler.
                # NOTE: we defer their creation because some users (e.g.,
                # BlockingConnection adapter) don't use the event loop and these
                # sockets would get reported as leaks
                self._r_interrupt, self._w_interrupt = self.get_interrupt_pair()
                self.add_handler(self._r_interrupt.fileno(),
                                 self.read_interrupt,
                                 READ)
                interrupt_sockets_created = True
            else:
                interrupt_sockets_created = False
        try:
            # Run event loop
            while not self._stopping:
                self.poll()
                self.process_timeouts()
        finally:
            # Unregister and close ioloop-interrupt socket pair
            if interrupt_sockets_created:
                with self._mutex:
                    self.remove_handler(self._r_interrupt.fileno())
                    self._r_interrupt.close()
                    self._r_interrupt = None
                    self._w_interrupt.close()
                    self._w_interrupt = None

    def stop(self):
        """Request exit from the ioloop."""

        LOGGER.debug('Stopping IOLoop')
        self._stopping = True

        with self._mutex:
            if self._w_interrupt is None:
                self.close()
                return

            try:
                # Send byte to interrupt the poll loop, use write() for
                # consitency.
                os.write(self._w_interrupt.fileno(), b'X')
            except OSError as err:
                if err.errno != errno.EWOULDBLOCK:
                    raise
            except Exception as err:
                # There's nothing sensible to do here, we'll exit the interrupt
                # loop after POLL_TIMEOUT secs in worst case anyway.
                LOGGER.warning("Failed to send ioloop interrupt: %s", err)
                raise
        self.close()

    def close(self):
        pass

    def poll(self, write_only=False):
        """Wait for events on interested filedescriptors.

        :param bool write_only: Passed through to the hadnlers to indicate
            that they should only process write events.
        """
        while True:
            try:
                read, write, error = select.select(self._fd_events[READ],
                                                   self._fd_events[WRITE],
                                                   self._fd_events[ERROR],
                                                   self.get_next_deadline())
                break
            except _SELECT_ERROR as error:
                if _get_select_errno(error) == errno.EINTR:
                    continue
                else:
                    raise

        # Build an event bit mask for each fileno we've recieved an event for

        fd_event_map = defaultdict(int)
        for fd_set, ev in zip((read, write, error), (READ, WRITE, ERROR)):
            for fileno in fd_set:
                fd_event_map[fileno] |= ev

        self._process_fd_events(fd_event_map, write_only)

    def _process_fd_events(self, fd_event_map, write_only):
        """ Processes the callbacks for each fileno we've recieved events.
            Before doing so we re-calculate the event mask based on what is
            currently set in case it has been changed under our feet by a
            previous callback. We also take a store a refernce to the
            fd_event_map in the class so that we can detect removal of an
            fileno during processing of another callback and not generate
            spurious callbacks on it.

            :param dict fd_event_map: Map of fds to events recieved on them.
        """

        self._processing_fd_event_map = fd_event_map

        for fileno in dictkeys(fd_event_map):
            if fileno not in fd_event_map:
                # the fileno has been removed from the map under our feet.
                continue

            events = fd_event_map[fileno]
            for ev in [READ, WRITE, ERROR]:
                if fileno not in self._fd_events[ev]:
                    events &= ~ev

            if events:
                handler = self._fd_handlers[fileno]
                handler(fileno, events, write_only=write_only)


class KQueuePoller(SelectPoller):
    """KQueuePoller works on BSD based systems and is faster than select"""

    def __init__(self):
        """Create an instance of the KQueuePoller

        :param int fileno: The file descriptor to check events for
        :param method handler: What is called when an event happens
        :param int events: The events to look for

        """
        self._kqueue = select.kqueue()
        super(KQueuePoller, self).__init__()

    def update_handler(self, fileno, events):
        """Set the events to the current events

        :param int fileno: The file descriptor
        :param int events: The event mask

        """

        kevents = list()
        if not events & READ:
            if fileno in self._fd_events[READ]:
                kevents.append(select.kevent(fileno,
                                             filter=select.KQ_FILTER_READ,
                                             flags=select.KQ_EV_DELETE))
        else:
            if fileno not in self._fd_events[READ]:
                kevents.append(select.kevent(fileno,
                                             filter=select.KQ_FILTER_READ,
                                             flags=select.KQ_EV_ADD))
        if not events & WRITE:
            if fileno in self._fd_events[WRITE]:
                kevents.append(select.kevent(fileno,
                                             filter=select.KQ_FILTER_WRITE,
                                             flags=select.KQ_EV_DELETE))
        else:
            if fileno not in self._fd_events[WRITE]:
                kevents.append(select.kevent(fileno,
                                             filter=select.KQ_FILTER_WRITE,
                                             flags=select.KQ_EV_ADD))
        for event in kevents:
            self._kqueue.control([event], 0)
        super(KQueuePoller, self).update_handler(fileno, events)

    def _map_event(self, kevent):
        """return the event type associated with a kevent object

        :param kevent kevent: a kevent object as returned by kqueue.control()

        """
        if kevent.filter == select.KQ_FILTER_READ:
            return READ
        elif kevent.filter == select.KQ_FILTER_WRITE:
            return WRITE
        elif kevent.flags & select.KQ_EV_ERROR:
            return ERROR

    def poll(self, write_only=False):
        """Check to see if the events that are cared about have fired.

        :param bool write_only: Don't look at self.events, just look to see if
            the adapter can write.

        """
        while True:
            try:
                kevents = self._kqueue.control(None, 1000,
                                               self.get_next_deadline())
                break
            except _SELECT_ERROR as error:
                if _get_select_errno(error) == errno.EINTR:
                    continue
                else:
                    raise

        fd_event_map = defaultdict(int)
        for event in kevents:
            fileno = event.ident
            fd_event_map[fileno] |= self._map_event(event)

        self._process_fd_events(fd_event_map, write_only)

    def close(self):
        self._kqueue.close()


class PollPoller(SelectPoller):
    """Poll works on Linux and can have better performance than EPoll in
    certain scenarios.  Both are faster than select.

    """
    POLL_TIMEOUT_MULT = 1000

    def __init__(self):
        """Create an instance of the KQueuePoller

        :param int fileno: The file descriptor to check events for
        :param method handler: What is called when an event happens
        :param int events: The events to look for

        """
        self._poll = self.create_poller()
        super(PollPoller, self).__init__()

    def create_poller(self):
        return select.poll()  # pylint: disable=E1101

    def close(self):
        self._poll.close()

    def add_handler(self, fileno, handler, events):
        """Add a file descriptor to the poll set

        :param int fileno: The file descriptor to check events for
        :param method handler: What is called when an event happens
        :param int events: The events to look for

        """
        self._poll.register(fileno, events)
        super(PollPoller, self).add_handler(fileno, handler, events)

    def update_handler(self, fileno, events):
        """Set the events to the current events

        :param int fileno: The file descriptor
        :param int events: The event mask

        """
        super(PollPoller, self).update_handler(fileno, events)
        self._poll.modify(fileno, events)

    def remove_handler(self, fileno):
        """Remove a fileno to the set

        :param int fileno: The file descriptor

        """
        super(PollPoller, self).remove_handler(fileno)
        self._poll.unregister(fileno)

    def poll(self, write_only=False):
        """Poll until the next timeout waiting for an event

        :param bool write_only: Only process write events

        """
        while True:
            try:
                events = self._poll.poll(self.get_next_deadline())
                break
            except _SELECT_ERROR as error:
                if _get_select_errno(error) == errno.EINTR:
                    continue
                else:
                    raise

        fd_event_map = defaultdict(int)
        for fileno, event in events:
            fd_event_map[fileno] |= event

        self._process_fd_events(fd_event_map, write_only)


class EPollPoller(PollPoller):
    """EPoll works on Linux and can have better performance than Poll in
    certain scenarios. Both are faster than select.

    """
    POLL_TIMEOUT_MULT = 1

    def create_poller(self):
        if not hasattr(self.__class__, '__instance__'):
            self.__class__.__instance__ = select.epoll()
        elif self.__class__.__instance__ is None:
            self.__class__.__instance__ = select.epoll()
        return self.__class__.__instance__

    def close(self):
        self._poll.close()
        self.__class__.__instance__ = None



import sys
import time
import copy
import signal
import logging
import asyncore
from threading import Thread
from multiprocessing import Process
from servicebus import utils

if sys.version_info < (3, 0):
    from Queue import Queue
else:
    from queue import Queue


class ServiceBus(object):
    def __init__(self, configuration, queue_len=100):
        self.configuration = configuration
        self.rpc_services = {}
        self.message_services = {}
        self.rpc_services_threads = {}
        self.message_services_threads = {}
        self.node_name = ""
        self.after_fork_hook = None
        self.on_exit_hook = None
        self.queue_len = queue_len

    def after_fork(self):
        if self.after_fork_hook:
            self.after_fork_hook()

    def set_after_fork_hook(self, hook):
        self.after_fork_hook = hook

    def set_on_exit_hook(self, hook):
        self.on_exit_hook = hook

    def set_node_name(self, name):
        self.node_name = name

    def add_rpc_service(self, category, name, service):
        key = "%s.%s" % (category, name)
        self.rpc_services[key] = service

    def add_message_services(self, category, name, service):
        key = "%s.%s" % (category, name)
        self.message_services[key] = service

    def _prepare_message_service_threads(self):
        for key, service in self.message_services.items():
            thread = ServiceRunner(service, self.queue_len)
            thread.start()
            self.message_services_threads[key] = thread

    def _prepare_rpc_service_threads(self):
        for key, service in self.rpc_services.items():
            thread = ServiceRunner(service, self.queue_len)
            thread.start()
            self.rpc_services_threads[key] = thread

    def lookup_rpc_service_thread(self, category, name):
        key = "%s.%s" % (category, name)
        return self.rpc_services_threads.get(key, None)

    def lookup_message_service_thread(self, category, name):
        key = "%s.%s" % (category, name)
        return self.message_services_threads.get(key, None)

    def run_services(self, join=True):
        processes = []
        for host in self.configuration.hosts:
            process = Process(target=self.run_server, args=(self.configuration, host))
            process.daemon = True
            process.start()
            processes.append(process)

        # wait for all process
        if join:
            for process in processes:
                process.join()
        return processes

    def run_server(self, configuration, host):
        self.running = True
        receiver = None
        self.after_fork()
        self.prepare_service_threads()

        def process_exit(signum, frame):
            self.running = False
            try:
                self.stop_service_threads()
                if self.on_exit_hook is not None:
                    logging.info('[Server %s]: Call on_exit hook' % host)
                    self.on_exit_hook()
            except Exception as e:
                logging.exception(e)
            logging.info("[Server %s]: Shutdown" % host)
            raise SystemExit("Terminate")

        signal.signal(signal.SIGTERM, process_exit)

        while self.running:
            receiver = None
            try:
                logging.info('[Server %s]: Build Server' % host)
                receiver = configuration.create_receiver(host)
                receiver.set_service_bus(self)
                receiver.bind_queue_to_exchange(configuration.queue_name(), configuration.exchange_name)
                logging.info('[Server %s]: Start Receive' % host)
                self.running = receiver.start_receive()
            except Exception as e:
                logging.exception(e)

            if self.running:
                logging.info('[Server %s]: Connection lost, wait 10 second to retry' % host)
                time.sleep(10)
                if receiver is not None:
                    receiver.join_watcher()
        self.stop_service_threads()
        try:
            if self.on_exit_hook is not None:
                logging.info('[Server %s]: Call on_exit hook' % host)
                self.on_exit_hook()
        except Exception as e:
            logging.exception(e)
        logging.info("[Server %s]: Shutdown" % host)

    def __force_close_sockets(self):
        utils.close_sockets()

    def prepare_service_threads(self):
        self._prepare_message_service_threads()
        self._prepare_rpc_service_threads()

    def stop_service_threads(self):
        logging.info("Stop Service Threads")
        all_threads = []
        for msg_thread in self.message_services_threads.values():
            all_threads.append(msg_thread)

        for rpc_thread in self.rpc_services_threads.values():
            all_threads.append(rpc_thread)

        for sthread in all_threads:
            sthread.stop()

        for sthread in all_threads:
            sthread.join()


class ServiceRunner(Thread):
    def __init__(self, service, queue_len=1):
        super(ServiceRunner, self).__init__()
        self.service = service
        self.queue = Queue(queue_len)
        self.is_background = self.is_background_service(service)
        self.background_threads = []

    def is_background_service(self, service):
        if hasattr(service, 'background'):
            return service.background
        return False

    def run(self):
        while True:
            msg_type, params = self.queue.get()
            if msg_type == "stop":
                self.queue.task_done()
                break
            # We copy the service object is for each message process.
            # This copy is for Background Service running. If we do not
            # copy this object, Background Service will return last received
            # message's data. So if there has 2 same service running in background
            # will cause a status report bug.
            service = copy.copy(self.service)
            if self.is_background:
                logging.info("Run in background")
                thread = Thread(target=self.run_service, args=(service, msg_type, params))
                thread.start()
                self.add_background_service(thread)
            else:
                self.run_service(service, msg_type, params)
            self.queue.task_done()
        self.wait_background_services()

    def run_service(self, service, msg_type, params):
        try:
            if msg_type == "message":
                service.on_message(*params)
            elif msg_type == "call":
                service.on_call(*params)
        except Exception as e:
            logging.exception(e)

        if len(params) > 0:
            request = params[0]
            request.get_sender().close()

    def on_message(self, request):
        self.queue.put(("message", (request,)))

    def on_call(self, request, response):
        self.queue.put(("call", (request, response)))

    def stop(self):
        self.queue.put(("stop", None))

    def add_background_service(self, thread):
        nthreads = [thread]
        for t in self.background_threads:
            if t.is_alive():
                nthreads.append(t)
        self.background_threads = nthreads

    def wait_background_services(self):
        if len(self.background_threads) > 0:
            logging.info("Wait for background service threads")
            for t in self.background_threads:
                t.join()



import logging
from servicebus.parser import XmlRequestGenerator, XmlResponseParser


class Sender(object):
    def __init__(self, configuration, smart_route=True):
        self.configuration = configuration
        self.exchange_name = configuration.exchange_name
        self.caller = None
        self.callers = None
        self.smart_route = smart_route

    def get_caller(self, reverse=False):
        if self.caller is None:
            self.caller = self.configuration.create_sender(reverse)
            self.caller.set_exchange(self.exchange_name)
        return self.caller

    def choose_caller(self, target, reverse=False):
        if self.smart_route:
            return self.smart_route_choose_caller(target, reverse)
        else:
            return self.simple_choose_caller(target, reverse)

    def simple_choose_caller(self, target, reverse=False):
        return self.get_caller(reverse)

    def smart_route_choose_caller(self, target, reverse=False):
        callers = self.get_callers()
        if reverse:
            callers.reverse()
        ret = None
        for caller in callers:
            if self._caller_ping(caller, target):
                return caller
        if len(callers) > 0:
            return callers[0]
        return ret

    def _caller_ping(self, caller, target, timeout=3):
        try:
            ret = caller.call(target, "PING", timeout)
            return ret == "PONG"
        except Exception:
            return False

    def get_callers(self):
        if self.callers is None:
            self.callers = self.configuration.create_senders()
            for caller in self.callers:
                caller.set_exchange(self.exchange_name)
        return self.callers

    def parse_target(self, target):
        parts = target.split(".")
        if len(parts) != 3:
            raise Exception("Target not validate")
        return parts

    def ping(self, target, timeout=3):
        caller = self.get_caller(True)
        ret = caller.call(target, "PING", timeout)
        return ret == "PONG"

    def ping_all(self, target, timeout=3):
        callers = self.get_callers()
        total = len(callers)
        success = 0
        for caller in callers:
            try:
                caller.set_exchange(self.exchange_name)
                ret = caller.call(target, "PING", timeout)
                if ret == "PONG":
                    success += 1
            except Exception:
                pass
        return success, total

    def call(self, target, params, timeout=300, reverse=False):
        target, category, service = self.parse_target(target)
        caller = self.choose_caller(target, reverse)
        if caller is None:
            raise Exception("Cannot connect to %s" % target)
        req_msg = XmlRequestGenerator(self.configuration, category, service, params)
        ret = caller.call(target, req_msg.to_xml(), timeout)
        resp_parser = XmlResponseParser()
        return resp_parser.parse(ret)

    def send(self, target, params):
        target, category, service = self.parse_target(target)
        caller = self.choose_caller(target)
        if caller is None:
            raise Exception("Cannot connect to %s" % target)
        req_msg = XmlRequestGenerator(self.configuration, category, service, params)
        caller.send(target, req_msg.to_xml())

    def close(self):
        if self.caller:
            self.caller.close()
            self.caller = None

        if self.callers:
            for caller in self.callers:
                try:
                    caller.close()
                except Exception:
                    pass
            self.callers = None



import uuid
import signal
import logging
from datetime import datetime
from servicebus import pika
from servicebus import utils
from servicebus.watcher import PingWatcher


class TimeoutException(Exception):
    pass


class RabbitMQMessageDriver(object):
    """docstring for AbstractMessageReceiver"""
    def __init__(self, host, port, username, password, ssl=False, socket_timeout=5, heartbeat_interval=60):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.ssl = ssl
        self.connection = None
        self.channel = None
        self.running = True
        self.connected = False
        self.socket_timeout = socket_timeout
        self.heartbeat_interval = heartbeat_interval

    def create_connection(self):
        connection = pika.BlockingConnection(pika.ConnectionParameters(
            self.host,
            self.port,
            credentials=pika.PlainCredentials(self.username, self.password),
            ssl=self.ssl,
            heartbeat_interval=self.heartbeat_interval,
            socket_timeout=self.socket_timeout
        ))
        self.connected = True
        return connection

    def declare_exchange(self, exchange_name, exchange_type='direct'):
        self.channel.exchange_declare(exchange=exchange_name, exchange_type=exchange_type)

    def close(self):
        if self.connected:
            self.connected = False
            self.connection.close()

    def bind_queue_to_exchange(self, queue_name, exchange_name, exchange_type='direct'):
        self.queue_name = queue_name
        self.declare_exchange(exchange_name, exchange_type)
        self.channel.queue_declare(queue=queue_name, durable=True)
        self.channel.queue_bind(exchange=exchange_name, queue=queue_name, routing_key=queue_name)

    def ensure_connection(self):
        if not self.connected:
            self.connection = self.create_connection()
            self.channel = self.connection.channel()


class AbstractReceiver(RabbitMQMessageDriver):
    # this mothod just receive one message.
    # only used by test.
    def receive_one(self):
        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(self.__on_receive, queue=self.queue_name)

    def loop(self):
        while self.connected:
            try:
                self.channel.start_consuming()
            except KeyboardInterrupt:
                self.running = False
                self.connected = False
                self.__safe_close()
            except EnvironmentError as e:
                if e.errno == 4 and not self.running:
                    logging.info("Got SIGTERM or SIGINT")
                    self.running = False
                    self.connected = False
                    self.__safe_close()
                else:
                    self.connected = False
                    self.__safe_close()
                    raise e
            except Exception as e:
                logging.exception(e)
                self.connected = False
                self.__safe_close()
                raise e
        if not self.connected:
            self.connected = False
            self.__safe_close()
        logging.info("All connection closed!")

    def __safe_close(self):
        try:
            self.channel.stop_consuming()
        except Exception:
            pass
        try:
            self.channel.close()
        except Exception:
            pass
        try:
            self.connection.close()
        except Exception:
            pass
        try:
            self.connection.close_ioloop()
        except Exception:
            pass

    def response_message(self, channel, method, header, message):
        channel.basic_publish(exchange='',
                              routing_key=header.reply_to,
                              properties=pika.BasicProperties(correlation_id=header.correlation_id),
                              body=message.encode())

    def on_rpc(self, channel, method, header, body):
        self.response_message(channel, method, header, None)

    def on_message(self, channel, method, header, body):
        pass

    def start_receive(self, count=None):
        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(self.__on_receive, queue=self.queue_name)
        self.watcher = PingWatcher.start_watch(self)
        try:
            self.loop()
            return self.running
        finally:
            self.watcher.stop()

    def process_exit(self, signum, frame):
        logging.info("Shutdown")
        self.running = False
        self.connected = False
        self.__safe_close()

    def join_watcher(self):
        try:
            if self.watcher:
                self.watcher.join()
        except Exception:
            pass

    def __on_receive(self, channel, method, header, body):
        try:
            channel.basic_ack(delivery_tag=method.delivery_tag)
            ebody = body.decode()
            if hasattr(header, 'reply_to') and header.reply_to is not None:
                # Here is a RPC call
                if ebody == "PING":
                    self.response_message(channel, method, header, "PONG")
                else:
                    self.on_rpc(channel, method, header, ebody)
            else:
                # Here is just send a message
                self.on_message(channel, method, header, ebody)
        except Exception as e:
            logging.exception(e)


class AbstractMessageSender(RabbitMQMessageDriver):
    def set_exchange(self, exchange_name, exchange_type='direct'):
        self.exchange_name = exchange_name
        self.declare_exchange(exchange_name, exchange_type)

    def send(self, target, msg):
        try:
            self.ensure_connection()
            self.channel.basic_publish(exchange=self.exchange_name, routing_key=str(target), body=msg)
        finally:
            self.close()


class MessageSender(AbstractMessageSender):
    def on_response(self, ch, method, props, body):
        if props.correlation_id == self.corr_id:
            self.response = body.decode()
            self.connection.remove_timeout(self.timeout_id)
            self.timeout_id = None
            self.channel.stop_consuming()

    def call(self, target, msg, timeout=300):
        self.ensure_connection()
        self.response = None
        self.timeout = False
        self.timeout_id = None
        try:
            self.corr_id = str(uuid.uuid4())
            result = self.channel.queue_declare(exclusive=True)
            callback_queue = result.method.queue
            self.channel.basic_publish(
                exchange=self.exchange_name,
                routing_key=str(target),
                properties=pika.BasicProperties(
                    reply_to=callback_queue,
                    correlation_id=self.corr_id,
                ),
                body=msg)
            self.channel.basic_consume(self.on_response, queue=callback_queue)

            def timeout_callback():
                self.channel.stop_consuming()
                self.timeout = True

            self.timeout_id = self.connection.add_timeout(timeout, timeout_callback)
            self.channel.start_consuming()

            if self.timeout:
                self.timeout_id = None
                raise Exception("Timeout")
            return self.response
        except Exception as e:
            logging.exception(e)
            raise e
        finally:
            if self.timeout_id:
                self.connection.remove_timeout(self.timeout_id)



import socket
import logging
import subprocess


def cmd(cmd_str):
    """
    Execute a command, return stdout and stderr in tuple
    """
    logging.debug("Execute Command: %s" % cmd_str)
    process = None
    try:
        process = subprocess.Popen(cmd_str, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        out, err = process.communicate()
        return out.decode('utf-8'), err.decode('utf-8')
    finally:
        if process is not None:
            process.stdout.close()
            process.stderr.close()


def cmd2(cmd_str):
    """
    Execute a command return retcode, stdout and stderr in tuple
    """
    logging.debug("Execute Command: %s" % cmd_str)
    process = None
    try:
        process = subprocess.Popen(cmd_str, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        out, err = process.communicate()
        ret_code = process.returncode
        return ret_code, out.decode('utf-8'), err.decode('utf-8')
    finally:
        if process is not None:
            process.stdout.close()
            process.stderr.close()


def get_host_name():
    return socket.gethostname()



class Event(object):
    def __init__(self, eid, category, service, token, params, version):
        self.id = eid
        self.category = category
        self.service = service
        self.token = token
        self.params = params
        self.version = version



"""
Connection Adapters
===================

Pika provides multiple adapters to connect to RabbitMQ:

- adapters.select_connection.SelectConnection: A native event based connection
  adapter that implements select, kqueue, poll and epoll.
- adapters.tornado_connection.TornadoConnection: Connection adapter for use
  with the Tornado web framework.
- adapters.blocking_connection.BlockingConnection: Enables blocking,
  synchronous operation on top of library for simple uses.
- adapters.twisted_connection.TwistedConnection: Connection adapter for use
  with the Twisted framework
- adapters.libev_connection.LibevConnection: Connection adapter for use
  with the libev event loop and employing nonblocking IO

"""
from servicebus.pika.adapters.base_connection import BaseConnection
from servicebus.pika.adapters.blocking_connection import BlockingConnection
from servicebus.pika.adapters.select_connection import SelectConnection
from servicebus.pika.adapters.select_connection import IOLoop

# Dynamically handle 3rd party library dependencies for optional imports

try:
    from servicebus.pika.adapters.tornado_connection import TornadoConnection
except ImportError:
    TornadoConnection = None

try:
    from servicebus.pika.adapters.twisted_connection import TwistedConnection
    from servicebus.pika.adapters.twisted_connection import TwistedProtocolConnection
except ImportError:
    TwistedConnection = None
    TwistedProtocolConnection = None

try:
    from servicebus.pika.adapters.libev_connection import LibevConnection
except ImportError:
    LibevConnection = None



from servicebus.sender import Sender


class Request(object):
    def __init__(self, event, receiver):
        self.event = event
        self.receiver = receiver
        self.sender = Sender(receiver.service_bus.configuration)

    def get_params(self):
        return self.event.params

    def get_event(self):
        return self.event

    def get_sender(self):
        return self.sender



import signal
import pickle
import struct
import logging
import logging.handlers
from multiprocessing import Process

try:
    from SocketServer import BaseRequestHandler, ThreadingUDPServer
except Exception:
    from socketserver import BaseRequestHandler, ThreadingUDPServer


DEFAULT_FORMAT_PATTERN = '[%(asctime)s PID:%(process)d]%(levelname)s:%(message)s'


class LogRecordHandler(BaseRequestHandler):
    def handle(self):
        data = self.request[0].strip()
        self.process_data(data)

    def process_data(self, data):
        chunk = data[0:4]
        if len(chunk) < 4:
            return
        slen = struct.unpack(">L", chunk)[0]
        chunk = data[4:slen + 4]
        try:
            obj = self.unpickle(chunk)
            record = logging.makeLogRecord(obj)
            self.handle_log_record(record)
        except Exception as e:
            logging.exception(e)

    def unpickle(self, data):
        return pickle.loads(data)

    def handle_log_record(self, record):
        logging.getLogger().handle(record)


class LoggingService(ThreadingUDPServer):
    max_packet_size = 16384

    @classmethod
    def init_logging(cls, port, level=logging.INFO):
        logger = logging.getLogger()
        shandler = logging.handlers.DatagramHandler('127.0.0.1', port)
        logger.addHandler(shandler)
        logger.setLevel(level)

    def __init__(self, file_name, port, level=logging.INFO,
                 format_pattern=DEFAULT_FORMAT_PATTERN):
        self.file_name = file_name
        self.level = level
        self.format_pattern = format_pattern
        ThreadingUDPServer.__init__(self, ('127.0.0.1', port), LogRecordHandler)
        self.abort = 0
        self.timeout = 1

    def serv_untile_stopped(self):
        import select
        abort = 0
        while not abort:
            rd, wr, ex = select.select([self.socket.fileno()],
                                       [], [],
                                       self.timeout)
            if rd:
                self.handle_request()
            abort = self.abort

    def configure_logger(self):
        self.logger = logging.getLogger()
        self.logger.setLevel(self.level)
        handler = logging.handlers.TimedRotatingFileHandler(
            filename=self.file_name, when='D', backupCount=15
        )
        handler.setFormatter(logging.Formatter(self.format_pattern))
        self.logger.addHandler(handler)

    @classmethod
    def run_in_process(cls, filename, port, after_fork_hook):
        if after_fork_hook:
            after_fork_hook()
        lserver = LoggingService(filename, port)
        lserver.configure_logger()
        logging.info("Start Logging Server")
        signal.signal(signal.SIGTERM, lserver.process_exit)
        lserver.serv_untile_stopped()

    def process_exit(self, signum, frame):
        self.abort = True

    @classmethod
    def start_logging_server(cls, filename, port, after_fork_hook):
        process = Process(target=cls.run_in_process, args=(filename, port, after_fork_hook))
        process.daemon = True
        process.start()
        return process



from distutils.core import setup

long_description = """Provide a basic RPC and Message process framework based on RabbitMQ message server.

Features:
      1. Auto reconnect when network is down
      2. Support SSL and dynamic token validation in message transfer
      3. Use multi-path to auto switch RabbitMQ server when major RabbitMQ server is down
"""

setup(
    name='py-servicebus',
    version='2.15',
    author='Rain Li',
    author_email='blacktear23@gmail.com',
    url='https://github.com/blacktear23/py-servicebus',
    download_url='https://github.com/blacktear23/py-servicebus/downloads',
    description='Provide a basic RPC and Message process framework based on RabbitMQ message server.',
    long_description=long_description,
    packages=['servicebus', 'servicebus.pika', 'servicebus.pika.adapters'],
    license='',
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Intended Audience :: Developers',
                 'Operating System :: OS Independent',
                 'License :: OSI Approved :: BSD License',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Topic :: Software Development :: Libraries',
                 ]
)



import logging
import threading
from servicebus.parser import XmlMessageParser, XmlResponseGenerator
from servicebus.message import AbstractReceiver
from servicebus.request import Request


class RPCResponse(object):
    def __init__(self, event, channel, method, header, receiver):
        self.event = event
        self.channel = channel
        self.method = method
        self.header = header
        self.receiver = receiver

    def send(self, message):
        msg = XmlResponseGenerator(self.event.id, message)
        self.receiver.response_message(self.channel, self.method, self.header,
                                       msg.to_xml())


class MessageBusReceiver(AbstractReceiver):
    def get_message_parser(self):
        return self.message_parser

    def set_service_bus(self, service_bus):
        self.service_bus = service_bus
        self.message_parser = XmlMessageParser()
        self.message_parser.set_configuration(service_bus.configuration)

    def on_rpc(self, channel, method, header, body):
        event = self.message_parser.parse(body)
        if not self.message_parser.validate_token(event.token):
            logging.error('Token Error')
            msg = XmlResponseGenerator(event.id, "Token not valid!")
            self.response_message(channel, method, header, msg.to_xml())
            return

        response = RPCResponse(event, channel, method, header, self)
        request = Request(event, self)
        service = self.service_bus.lookup_rpc_service_thread(event.category, event.service)

        if service is None:
            error_msg = 'Cannot Find RPC Service: %s.%s' % (event.category, event.service)
            logging.error(error_msg)
            msg = XmlResponseGenerator(event.id, error_msg)
            self.response_message(channel, method, header, msg.to_xml())
        else:
            logging.info("Call RPC Service %s.%s" % (event.category, event.service))
            service.on_call(request, response)

    def on_message(self, channel, method, header, body):
        event = self.message_parser.parse(body)
        if not self.message_parser.validate_token(event.token):
            logging.error('Token Error')
            return

        request = Request(event, self)
        service = self.service_bus.lookup_message_service_thread(event.category, event.service)

        if service is None:
            error_msg = 'Cannot Find Message Service: %s.%s' % (event.category, event.service)
            logging.error(error_msg)
        else:
            logging.info("Call Message Service %s.%s" % (event.category, event.service))
            service.on_message(request)



import time
import logging
from servicebus.message import MessageSender
from servicebus.receiver import MessageBusReceiver
from servicebus.command import get_host_name


DEFAULT_EXCHANGE_NAME = 'py-servicebus'


class Configuration(object):
    """
    params: config
    config should be a hash:
    config['hosts']     = ['host1', 'host2']
    config['port']      = port default is 5672
    config['ssl_port']  = ssl port default is 5671
    config['user']      = user name
    config['password']  = password
    config['use_ssl']   = True or False default is False
    config['node_name'] = node name aka queue name if not set default
                          is host name.
    config['exchange_name'] = RabbitMQ Exchange name or use default:
                                py-servicebus
    config['secret_token']  = Message token secret token seed
    """
    def __init__(self, config):
        self.hosts = config['hosts']
        self.secret_token = config['secret_token']
        self.port = 5672
        if 'port' in config:
            self.port = int(config['port'])
        self.ssl_port = 5671
        if 'ssl_port' in config:
            self.ssl_port = int(config['ssl_port'])
        self.use_ssl = False
        if 'use_ssl' in config:
            self.use_ssl = config['use_ssl']
        self.user = None
        if 'user' in config:
            self.user = config['user']
        self.password = None
        if 'password' in config:
            self.password = config['password']
        self.node_name = get_host_name()
        if 'node_name' in config:
            self.node_name = config['node_name']
        self.exchange_name = DEFAULT_EXCHANGE_NAME
        if 'exchange_name' in config:
            self.exchange_name = config['exchange_name']
        self.socket_timeout = 5
        if 'socket_timeout' in config:
            self.socket_timeout = config['socket_timeout']
        self.heartbeat_interval = 60
        if 'heartbeat_interval' in config:
            self.heartbeat_interval = config['heartbeat_interval']

    """
    Thie method will create a message receiver.
    params: host    RabbitMQ server's host
    """
    def create_receiver(self, host, close_ioloop=True):
        receiver = MessageBusReceiver(
            host,
            self.get_port(),
            self.user,
            self.password,
            self.use_ssl,
            self.socket_timeout,
            self.heartbeat_interval
        )
        try:
            receiver.ensure_connection()
        except Exception as e:
            if close_ioloop:
                try:
                    receiver.connection.close_ioloop()
                except Exception:
                    pass
            raise e
        return receiver

    """
    This method will get an availiable host to send message to.
    """
    def create_sender(self, reverse=False):
        for i in range(3):
            caller = self.__get_availiable_message_sender(reverse)
            if caller is not None:
                break
            time.sleep(1)

        if not caller:
            raise Exception("Cannot Connect to Message Queue!")
        return caller

    """
    This method will create N senders N is equals length of host.
    Each of sender is connect to each host.
    If host cannot connected, return value will not contains this host's sender.
    """
    def create_senders(self):
        ret = []
        for host in self.hosts:
            try:
                caller = self.__create_message_sender(host)
                caller.ensure_connection()
                ret.append(caller)
            except Exception as e:
                logging.exception(e)
        return ret

    def queue_name(self):
        return self.node_name

    def get_port(self):
        if self.use_ssl:
            port = self.ssl_port
        else:
            port = self.port
        return port

    def __get_availiable_message_sender(self, reverse):
        hosts = self.hosts[:]
        if reverse:
            hosts.reverse()
        for host in hosts:
            try:
                caller = self.__create_message_sender(host)
                caller.ensure_connection()
                return caller
            except Exception as e:
                logging.exception(e)
        return None

    def __create_message_sender(self, host):
        caller = MessageSender(
            host,
            self.get_port(),
            self.user,
            self.password,
            self.use_ssl,
            self.socket_timeout
        )
        return caller
